# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 10 Feb 2021: Initial Coding

"""
Timing, timer, time and suchlike tools.
"""


__all__ = ['Stopwatch', 'timestamp']


from collections import deque
from contextlib import AbstractContextManager
from datetime import datetime
from time import time


class _StopwatchPause(AbstractContextManager):
    """
    A context manager for measuring stopwatch pauses.
    """
    __slots__ = ('parent')

    def __init__(self, stopwatch):
        """
        Measures the duration of pauses in a running stopwatch.

        Parameters
        ----------
        stopwatch : Stopwatch
            The parent object.
        """
        self.parent = stopwatch

    def __exit__(self, *args):
        """
        Stops the pause and updates the parent :py:class:`Stopwatch`.
        """
        self.parent.unpause()
        return super().__exit__(*args)


class Stopwatch(AbstractContextManager):
    """
    Rough lightweight timer context manager.

    This is not intended to be used for precise benchmarking, but can
    give a good idea of how long operations in a `with` block take.

    The context manager can be reused as many times as necessary. When
    first created, it reports time relative to its creation time,
    unless specifically requested otherwise.

    .. py:attribute:: start

       The start time of the stopwatch, as a floating point timestamp
       from Epoch (see :py:func:`time.time`). This is valid regardless
       of whether the stopwatch is running or not.

    .. py:attribute:: end

       The end time of the stopwatch, as a floating point timestamp
       from Epoch (see :py:func:`time.time`). If the stopwatch is
       running, this is set to None. If it is stopped or paused, this
       attribute records the time of the stop or pause.

    .. py:attribute:: pauses

       A sequence containing ``(start, end)`` tuples for every pause
       triggered since the last restart. The last element may be a
       placeholder object if the stopwatch is paused. All tuple
       elements have units of seconds from Epoch (see
       :py:func:`time.time`).

    .. py:attribute:: pause_duration

       The sum of the durations of all the pauses, in seconds.

    This class is not thread safe.
    """
    __slots__ = ('start', 'end', 'pauses', 'pause_duration')

    def __init__(self, start=None):
        """
        Create a new stopwatch, starting from now.

        Parameters
        ----------
        start : float or None, optional
            The start timestamp, or now if None. The default is None.
        """
        # The container must be created before it is cleared
        self.pauses = deque()
        # This clears the pauses container and sets pause_duration
        self.restart()
        if start is not None:
            self.start = start

    def __enter__(self):
        """
        Restart the timer.

        See :py:meth:`restart`.
        """
        self.restart()
        return super().__enter__()

    def __exit__(self, *args):
        """
        Stops the timer.

        See :py:meth:`stop`.
        """
        self.stop()
        return super().__exit__(*args)

    def __str__(self):
        """
        Pretty-prints the duration of this stopwatch, with a label to
        indicate if it is running.

        Returns
        -------
        str :
            A string representation of the stopwatch duration.
        """
        df = self.duration
        di = int(df)
        s = 'Running' if self.end is None else 'Stopped'
        return f'{di // 3600:02d}:{(di // 60) % 60:02d}:{df % 60:0.2f} [{s}]'

    @property
    def duration(self):
        """
        The duration of the stopwatch timer, including any intervening
        pauses. If the stopwatch is running, this is the duration until
        now. If stopped, this is the duration between start and stop.
        """
        start = self.start + self.pause_duration
        end = time() if self.end is None else self.end
        return end - start

    def restart(self):
        """
        Clears the pause sequence and restarts the timer.
        """
        self.clear()
        self.end = None
        self.start = time()

    def pause(self):
        """
        Pause the stopwatch, if not already paused.

        This method returns a context manager. Entering the context
        manager does nothing, but exiting it unpauses this timer.

        Calling this method multiple times without unpausing will
        lead to potentially unexpected behavior. All context managers
        returned by this method unpause the stopwatch. That means, for
        example, that nesting context managed calls to :py:meth:`pause`
        will unpause when the innermost context manager exits, not the
        outermost.

        Returns
        -------
        A subsidiary context manager that can be used to automatically
        :py:meth:`unpause` when it exits.
        """
        if self.end is None:
            self.end = time()
        return _StopwatchPause(self)

    def unpause(self):
        """
        Unpause the timer, if it has been paused.

        :py:attr:`pause_duration` is updated and a new entry is
        appended to :py:attr:`pauses` if the timer was not running.
        """
        self._unstop()
        self.end = None

    def stop(self):
        """
        Stops the timer.

        Calling this method multiple times will update :py:attr:`end`
        and :py:attr:`pause_duration`, and append a new entry to the
        :py:attr:`pauses` sequence from the last stop/pause.
        """
        self.end = self._unstop()

    def clear(self):
        """
        Removes all pause records, including any current ones.

        Clearing the records will affect the reported duration, whether
        the stopwatch is running or not.
        """
        self.pauses.clear()
        self.pause_duration = 0.0

    def _unstop(self):
        """
        If :py:attr:`end` is not None, update :py:attr:`pauses` and
        :py:attr:`pause_duration`.

        Used by :py:attr:`stop` and :py:attr:`unpause` to update the
        timer.

        Returns
        -------
        end : float
            The current time (at the start of the method).
        """
        end = time()
        start = self.end
        if start is not None:
            self.pauses.append((start, end))
            self.pause_duration += end - start
        return end


def timestamp(t=None):
    """
    Return the current or other date and time in the format
    ``YYYYMMDD_HHMMSS``.

    Parameters
    ----------
    t : datetime.datetime or None
        The date to format. If `None`, use the result of
        :py:meth:`datetime.datetime.now`.

    Returns
    -------
    str
        The formatted date.
    """
    if t is None:
        t = datetime.now()
    return t.strftime('%Y%m%d_%H%M%S')



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2024  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for working with PyQt6.

Utilities are only available if PyQt6 is present on the system. Some
attempt is made to support PyQt5 as well, but may be more sporadic.
"""


__all__ = []


try:
    from PyQt6.QtGui import QPixmap, QResizeEvent
    from PyQt6.QtWidgets import QLabel
except ImportError:
    from .. import _display_missing_extra
    _display_missing_extra('qt', 'PyQt5 or PyQt6')
    qt_enabled = False
else:
    __all__.extend(['ImageAspectLabel'])
    qt_enabled = True


if qt_enabled:
    class ImageAspectLabel(QLabel):
        """
        A label for holding a `QPixmap` image without altering the
        aspect ratio as the label resizes.

        The interface is identical to a normal `QLabel` in all other
        respects.
        """
        def __init__(self):
            super().__init__()
            self.pixmap_width = 1
            self.pixmap_height = 1

        def setPixmap(self, pm: QPixmap):
            if pm is None:
                self.pixmap_width = self.pixmap_height = 0
            else:
                self.pixmap_width = pm.width()
                self.pixmap_height = pm.height()
            self.updateMargins()
            super().setPixmap(pm)

        def resizeEvent(self, a0: QResizeEvent):
            self.updateMargins()
            super().resizeEvent(a0)

        def updateMargins(self):
            if self.pixmap_width <= 0 or self.pixmap_height <= 0:
                self.setContentsMargins(0, 0, 0, 0)
                return
            width, height = self.width(), self.height()
            if width <= 0 or height <= 0:
                return

            if width * self.pixmap_height > height * self.pixmap_width:
                pad = int(width - height * self.pixmap_width / self.pixmap_height + 0.5)
                self.setContentsMargins(pad // 2, 0, pad // 2 + pad % 2, 0)
            else:
                pad = int(height - width * self.pixmap_height / self.pixmap_width + 0.5)
                self.setContentsMargins(0, pad // 2, 0, pad // 2 + pad % 2)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for processing bits of latex using matplotlib for rendering.

The `matplotlib`_ setup routines are based on
http://stackoverflow.com/a/38008501/2988730.

This module relies on the ``[plot]`` and ``[latex]``
:ref:`extras <installation-extras>`. The :py:func:`render_latex` function
does not rely on ``[plot]``. All the other functions work with
`matplotlib`_. See :py:data:`haggis.mpl_util.plot_enabled`.


.. include:: /link-defs.rst
"""

__all__ = [
    'add_use_package', 'render_latex', 'package_list',
    'latex_exe', 'dvips_exe', 'convert_exe',
]


import tempfile, subprocess, io
from os import path


#: The list of packages loaded into the LaTeX preamble when this module
#: is imported. This list gets updated with every call to
#: :py:func:`add_use_package`. This allows :py:func:`setup_mpl` to work
#: correcly even when backends are changed under the ``[plot]`` extra.
package_list = ['amsmath', 'color', 'dashrule']  # Candidates: amssym, tikz


#: The name of the :program:`latex` executable. Either a full path, or a
#: program that the shell can find on the :envvar:`PATH` is necessary.
latex_exe = 'latex'


#: The name of the :program:`dvips` executable. Either a full path, or a
#: program that the shell can find on the :envvar:`PATH` is necessary.
dvips_exe = 'dvips'


#: The name of the `ImageMagick`_ :program:`convert` executable. Either
#: a full path, or a program that the shell can find on the
#: :envvar:`PATH` is necessary.
convert_exe = 'convert'


def add_use_package(package_name):
    r"""
    Add a single package via ``\usepackage{package_name}`` to the
    LaTeX premble.
    """
    if package_name not in package_list:
        package_list.append(package_name)


def render_latex(formula, file=None, format='png', *, fontsize=12, dpi=None,
                 transparent=False, bgcolor='white', fgcolor='black'):
    """
    Render a simple LaTeX formula into an image using external programs.

    If `file` is :py:obj:`None` (the default), return a
    :py:class:`~io.BytesIO` object containing the rendered image in PNG
    format. The stream is rewound, so can be read immediately.
    Otherwise, output to the specified file (which may be a file name
    string or any file-like object).

    The sequence of system commands run by this function is based
    largely on :program:`text2im` (http://www.nought.de/tex2im.php).
    """
    preamble = '\n'.join(r'\usepackage{%s}' % pkg for pkg in package_list)
    content = '\n'.join([
        r'\documentclass[{fontsize}pt]{{article}}',
        r'{preamble}',
        r'\usepackage[dvips]{{graphicx}}',
        r'\pagestyle{{empty}}',
        r'\pagecolor{{{bgcolor}}}',
        #r'\DeclareMathSizes{{{fontsize}}}{{{fontsize}}}{{{fontsize2}}}{{{fontsize3}}}',
        r'\begin{{document}}',
        r'{{\color{{{fgcolor}}}',
        r'$$ {formula} $$',
        r'}}\end{{document}}'
    ])
    content = content.format(
        preamble=preamble, formula=formula, bgcolor=bgcolor, fgcolor=fgcolor,
        fontsize=fontsize#, fontsize2=fontsize-2, fontsize3=fontsize-4
    )
    latex = [
        latex_exe, '-output-format=dvi', '-halt-on-error', #'-interaction=batchmode' 
    ]
    dvips = [dvips_exe, '-o', '-']
    convert = [
        convert_exe, '+adjoin', '-antialias'
    ]

    if dpi:
        convert.extend(['-density', '{0}x{0}'.format(dpi)])
    if transparent:
        convert.extend(['-transparent', bgcolor])

    rewind = False
    convert.append('-')                           # Input from pipe
    if isinstance(file, str):
        convert.append('{}:{}'.format(format, file))  # Output to file
        stdout = None
    else:
        if file is None:
            file = io.BytesIO()
            rewind = True
        convert.append(format + ':-')             # Output to pipe
        stdout = file

    with tempfile.TemporaryDirectory() as tmpdir:
        latex.append('-output-directory=' + tmpdir)
        proc1 = subprocess.Popen(latex, stdin=subprocess.PIPE,
                                 stdout=subprocess.DEVNULL,
                                 universal_newlines=True)
        proc1.communicate(input=content)

        dvips.extend(['-E', path.join(tmpdir, 'texput.dvi')])
        proc2 = subprocess.Popen(dvips, stdin=None, stdout=subprocess.PIPE,
                                 stderr=subprocess.DEVNULL)
        proc3 = subprocess.Popen(convert, stdin=proc2.stdout, stdout=stdout)
        proc2.stdout.close()
        proc3.communicate()

    if rewind:
        file.seek(0)
    return file


# Do this first to verify the [plot] extra
from . import mpl_util
if mpl_util.plot_enabled:
    import matplotlib


    __all__.extend([
        'setup_mpl', 'render_latex_mpl', 'pgf_tex_system',
    ])


    #: The name of the LaTeX PGF system, obtained from::
    #:
    #:    latex -v
    pgf_tex_system = 'pdftex'


    def setup_mpl():
        """
        Sets up the packages that should be used by `matplotlib`\ 's
        LaTeX processor.

        In addition to ensuring that some basic packages are imported in
        the preamble, this method enables latex usage in `matplotlib`_
        text elements such as titles and axis labels.

        This method must be called manually by any package wishing to
        use :py:func:`render_latex_mpl` with full capabilities. It
        should be used when the backend is changed to and from ``'pgf'``
        as well.


        .. include:: /link-defs.rst
        """
        # See http://matplotlib.org/users/usetex.html for details on text keys,
        # and http://matplotlib.org/users/pgf.html for details on PGF backend
        if matplotlib.get_backend() == 'pgf':
            from matplotlib.backends.backend_pgf import FigureCanvasPgf
            matplotlib.backend_bases.register_backend('pdf', FigureCanvasPgf)
            matplotlib.backend_bases.register_backend('png', FigureCanvasPgf)
            matplotlib.rc('pgf', texsystem=pgf_tex_system)
        else:
            matplotlib.rc('text', usetex=True)

        for package in package_list:
            add_use_package(package)


    def add_use_package(package_name):
        r"""
        Add a single package via ``\usepackage{package_name}`` to the
        list of `matplotlib`_\ 's LaTeX imports.

        Imports can be added to ``text.latex.preamble`` or
        ``pgf.preamble`` RC keys, depending on the current backend.
        Packages are also automatically added to the preamble of non-MPL
        rendered LaTeX.


        .. include:: /link-defs.rst
        """
        if package_name not in package_list:
            package_list.append(package_name)

        def set_preamble(which):
            preamble = matplotlib.rcParams.setdefault(which, [])
            use_str = r'\usepackage{%s}' % package_name
            if use_str not in preamble:
                preamble.append(use_str)

        if matplotlib.get_backend() == 'pgf':
            set_preamble('pgf.preamble')
        else:
            set_preamble('text.latex.preamble')


    def render_latex_mpl(formula, file=None, fontsize=12, **kwargs):
        """
        Render a simple LaTeX formula into an image using
        `matplotlib`_ figures.

        If `file` is :py:obj:`None` (the default), returns a
        :py:class:`~io.BytesIO` object containing the rendered image in
        the specified format. The stream is rewound, so can be read
        immediately. Otherwise outputs to the specified file (which may
        be a file name string or any file-like object).

        All arguments besides `file` and `fontsize` are passed through to
        :py:meth:`matplotlib.figure.Figure.savefig`.

        This method is based on the following Stack Overflow answer:
        http://stackoverflow.com/a/31371907/2988730


        .. include:: /link-defs.rst
        """
        with mpl_util.figure_context(figsize=(0.01, 0.01)) as fig:
            fig.text(0, 0, u'${}$'.format(formula), fontsize=fontsize,
                     horizontalalignment='left', verticalalignment='bottom')
            fig.frameon = False
            return mpl_util.save_figure(fig, file, bbox_inches='tight')


if __name__ == '__main__':
    if mpl_util.plot_enabled:
        setup_mpl()
    eqn = (r'P_k = \frac{\sqrt{\frac{1}{N - 1}\sum_{i=1}^{N} '
                                    r'(x_i - \bar{x})^2}}{\bar{x}}')
    render_latex(eqn, file='renderer_demo.png')
    if mpl_util.plot_enabled:
        render_latex_mpl(eqn, format='png', file='renderer_demo_mpl.png')

    mat = (r'\begin{bmatrix} '
           r'a_{11} & b_{12} & \dots  & z_{1n} \\ '
           r'a_{21} & b_{22} & \dots  & z_{2n} \\ '
           r'\vdots & \vdots & \ddots & \vdots \\ '
           r'a_{n1} & a_{n2} & \dots  & z_{nn} '
           r'\end{bmatrix}')
    render_latex(mat, file='renderer_demo2.png')
    if mpl_util.plot_enabled:
        render_latex_mpl(mat, format='png', file='renderer_demo2_mpl.png')

    text = (r'\textcolor{blue}{ -\bullet- \; }')
    render_latex(text, file='renderer_demo3.png')
    if mpl_util.plot_enabled:
        render_latex_mpl(text, format='png', file='renderer_demo3_mpl.png')




# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for processing CSV files.

Among other things, this module registers a 'text' dialect with the
built-in :py:mod:`csv` module, suitable for ingesting plain text
arranged in space-separated colmns.
"""


__all__ = ['load_as_numbers', 'load_as_columns', 'reformat']


import csv, contextlib

import numpy


csv.register_dialect('text', skipinitialspace=True, delimiter=' ')


def reformat(data, format='normal'):
    """
    Convert a normal CSV dataset stored by rows into a different format.

    Valid formats are

      - ``'normal'``: Return as a list of rows.
      - ``'transpose'``: Return a list of columns.
      - ``'numpy'``: Return a numpy array.
    """
    if format not in reformat._formats:
        raise ValueError('Unsupported format {}'.format(format))
    formatter = reformat._formats[format]
    if formatter is None:
        return data
    return formatter(data)


reformat._formats = {
    'normal': None,
    'transpose': lambda data: list(map(list, zip(*data))),
    'numpy': numpy.array,
}


def load_as_numbers(file, header_lines=0, dialect='text',
                    format='normal', empty=float('nan'), **kwargs):
    """
    Load a CSV file as a numbers.

    Parameters
    ----------
    file : str or file-like
        Strings are assumed to be file names and opened. Other file-like
        objects are not closed when this function returns.
    header_lines : int
        The number of lines to skip from the beginning of the file.
    dialect : str or csv.Dialect
        The dialect to use. String options can be obtained from
        :py:func:`csv.list_dialects`.
    format : str
        One of the following data formats to use for the return value:

          - ``'normal'``: Return as a list of rows.
          - ``'transpose'``: Return a list of columns.
          - ``'numpy'``: Return a numpy array.
    empty : number
        The value to use for empty strings.
    kwargs : dict
        Any additional parameters to pass to :py:func:`csv.reader`.

    Raises
    ------
    ValueError
        If any of the elements of the file can not be converted to a
        :py:class:`float` or :py:class:`int`.
    """
    if isinstance(file, str):
        mgr = file = open(file, 'r')
    else:
        mgr = contextlib.ExitStack()

    # This is not an efficient way to load the file. Will not work well for
    # large files
    lines = []
    with mgr:
        reader = csv.reader(file, dialect, **kwargs)
        for cnt, line in enumerate(reader):
            if cnt < header_lines:
                continue
            for col, item in enumerate(line):
                line[col] = float(item) if item else empty
            lines.append(line)

    return reformat(lines, format)


def load_as_columns(file, header_lines=0, empty=float('nan'),
                    dialect='text', **kwargs):
    """
    Load a CSV file as sequence of columns rather than rows.

    Parameters
    ----------
    file : str or file-like
        Strings are assumed to be file names and opened. Other file-like
        objects are not closed when this function returns.
    header_lines : int
        The number of lines to skip from the beginning of the file.
    dialect : str or csv.Dialect
        The dialect to use. String options can be obtained from
        :py:func:`csv.list_dialects`.
    empty : number
        The value to use for missing elements. If :py:obj:`None`, the
        data may not be ragged: each line must contain the same number
        of fields as the first.
    kwargs : dict
        Any additional parameters to pass to :py:func:`csv.reader`.

    Raises
    ------
    ValueError
        If the file contains a ragged array and `empty` is set to
        :py:obj:`None`.
    """
    if isinstance(file, str):
        mgr = file = open(file, 'r')
    else:
        mgr = contextlib.ExitStack()

    # This is not an efficient way to load the file. Will not work well for
    # large files
    columns = []
    with mgr:
        reader = csv.reader(file, dialect, **kwargs)
        for cnt, line in enumerate(reader):
            if cnt < header_lines:
                continue
            elif cnt == header_lines:
                columns = [[item] for item in line]
                continue

            if len(columns) != len(line):
                if empty is None:
                    raise ValueError('Number of items in line {} '
                                     'does not match first line '
                                     '(line {})'.format(cnt, header_lines))
                if len(columns) < len(line):
                    columns.append([empty] * len(columns[0]))
                else:
                    line.append([empty] * (len(columns) - len(line)))

            for col, item in zip(columns, line):
                col.append(item)

    return columns



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 12 Jan 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.math` module.
"""


import numpy
from numpy.testing import assert_allclose, assert_array_equal
from pytest import raises

from ..math import segment_distance, real_divide
from .util import plotting_context, save


class TestSegmentDistance:
    @staticmethod
    def point(t, p1, p2):
        return p1 + t * numpy.subtract(p2, p1)

    def test_one(self, plots):
        p = [0, 3]
        p1 = [-numpy.sqrt(3), 0]
        p2 = [0, 1]
        dist_seg, t_seg = segment_distance(p, p1, p2, return_t=True, segment=True)
        dist_line, t_line = segment_distance(p, p1, p2, return_t=True, segment=False)

        assert_allclose(dist_seg, 2.0, atol=1e-15, rtol=0.0)
        assert_allclose(dist_line, numpy.sqrt(3), atol=1e-15, rtol=0.0)
        assert_allclose(t_seg, 1.5, atol=1e-15, rtol=0.0)
        assert t_seg == t_line

        if plots:
            bounds = numpy.stack((p1, p2), axis=-1)
            p = numpy.array(p)
            pn = self.point(t_line, p1, p2)
            with plotting_context(constrained_layout=True) as fig:
                ax = fig.subplots(1)

                # Draw line
                ax.axline(*bounds.T, color='k', linestyle='--')
                ax.plot(*bounds, color='k', ls='-')

                # Draw normals
                ax.plot(*numpy.stack((p, pn), -1),
                        color='k', linestyle='--')
                ax.plot(*numpy.stack((p, p2), -1), color='k', linestyle='--')

                # Draw markers
                ax.plot(*p, color='k', linestyle='none',
                        marker='o', markersize=10)
                ax.plot(*bounds, color='k', linestyle='none',
                        marker='o', markersize=10)
                ax.plot(*pn, color='k', linestyle='none',
                        marker='o', markerfacecolor='w', markersize=10)

                # Annotate inputs
                ax.text(*p + [-0.15, 0], r'$\vec{p}$',
                        ha='right')
                ax.text(*bounds[:, 0] + [-0.15, 0.1], r'$\vec{p}_1$',
                        ha='right')
                ax.text(*bounds[:, 1] + [-0.15, 0.1], r'$\vec{p}_2$',
                        ha='right')

                # Annotate norms
                ax.text(*0.5 * (p + p2) + [-0.15, 0],
                        fr'$\Delta_1 = {dist_seg:0.0f}$',
                        va='center', ha='center', rotation=90)
                ax.text(*0.5 * (p + pn) + [0.15, 0.1],
                        fr'$\Delta_2 = \sqrt{{{dist_line**2:0.0f}}}$',
                        va='center', ha='center',
                        rotation=numpy.rad2deg(numpy.arctan(-t_line)))

                # Annotate projections
                ax.text(*pn + [0.12, -0.15], f'$t = {t_seg:0.1f}$')

                ax.axis('equal')
                save(fig, __name__, 'one')

    def test_n_points(self, plots):
        p = [[0, 3],
             [-0.5 * (numpy.sqrt(3) + 1), 0.5 * (numpy.sqrt(3) + 1)],
             [1, 1 - numpy.sqrt(3)]]
        p1 = [-numpy.sqrt(3), 0]
        p2 = [0, 1]
        dist_seg, t_seg = segment_distance(p, p1, p2, axis=1, return_t=True,
                                           segment=True)
        dist_line, t_line = segment_distance(p, p1, p2, axis=-1, return_t=True,
                                             segment=False)

        assert_allclose(dist_seg, [2.0, 1.0, 2.0], atol=1e-15, rtol=0.0)
        assert_allclose(dist_line, [numpy.sqrt(3), 1.0, 2.0],
                        atol=1e-15, rtol=0.0)
        assert_allclose(t_seg, [1.5, 0.5, 1.0], atol=1e-15, rtol=0.0)
        assert_array_equal(t_seg, t_line)

        if plots:
            def annotated_line(ax, start, end, label, dist, angle=-60):
                ax.plot([start[0], end[0]], [start[1], end[1]], c='k', ls='--')
                dist = f'{dist:0.0f}' if numpy.isclose(dist, round(dist)) \
                                 else fr'\sqrt{{{dist**2:0.0f}}}'
                ax.text(*0.5 * (start + end) + [0.15, 0.1],
                        fr'$\Delta_{{{label}}} = {dist}$',
                        va='center', ha='center', rotation=angle)

            p = numpy.array(p)
            bounds = numpy.stack((p1, p2), axis=-1)

            with plotting_context(constrained_layout=True) as fig:
                ax = fig.subplots(1)

                # Plot segment
                ax.axline(*bounds.T, color='k', linestyle='--')
                ax.plot(*bounds, color='k', ls='-')

                for i, s, d1, d2, t in zip(
                                 'ABC', p, dist_seg, dist_line, t_line):
                    e1 = self.point(t, p1, p2)

                    # Plot & Label Normals
                    if d1 == d2:
                        annotated_line(ax, s, e1, i, d2)
                    else:
                        e2 = p1 if t < 0 else p2
                        annotated_line(ax, s, e2, f'{i}_2', d2, -90)
                        annotated_line(ax, s, e1, f'{i}_1', d1)

                    # Annotate inputs
                    ax.text(*s + [-0.15, 0], fr'$\vec{{p}}_{{{i}}}$',
                            ha='right')

                    # Annotate projections
                    if not numpy.isclose(t, 0) and not numpy.isclose(t, 1):
                        ax.text(*e1 + [0.12, -0.15], f'$t = {t:0.1f}$')

                # Draw markers
                ax.plot(*p.T, color='k', linestyle='none',
                        marker='o', markersize=10)
                ax.plot(*self.point(t_line[:, None], p1, p2).T,
                        color='k', linestyle='none', marker='o',
                        markerfacecolor='w', markersize=10)
                ax.plot(*bounds, color='k', linestyle='none',
                        marker='o', markersize=10)

                # Annotate Inputs
                ax.text(*bounds[:, 0] + [-0.15, 0.1], r'$\vec{p}_1$',
                        ha='right')
                ax.text(*bounds[:, 1] + [-0.15, 0.1], r'$\vec{p}_2$',
                        ha='right')

                ax.axis('equal')
                save(fig, __name__, 'n_points')

    def test_n_lines(self, plots):
        p = [1.25, 1.25]
        p1 = [[0.5, 2.5], [2, 1.5], [2.5, -0.5], [1, -2],
              [-1.5, -1.5], [-2, -1], [-2, 1], [-1, 2]]
        p2 = [[2.5, 0.5], [2, -1.5], [1, -2], [-1, -2],
              [-2, -1], [-2, 1], [-1, 2], [1.5, 2]]

        dist_seg, t_seg = segment_distance(p, p1, p2, axis=1, return_t=True,
                                           segment=True)
        dist_line, t_line = segment_distance(p, p1, p2, axis=-1, return_t=True,
                                             segment=False)

        sh = numpy.sqrt(1 / 2)
        hsh = sh / 2
        qs170 = numpy.sqrt(170) / 4

        assert_allclose(dist_seg, [hsh, 3 / 4, numpy.sqrt(74) / 4, qs170,
                                   11 * hsh, qs170, 3 * numpy.sqrt(10) / 4,
                                   3 / 4], atol=1e-15, rtol=0.0)
        assert_allclose(dist_line, [hsh, 3 / 4, 3 * sh, 13 / 4,
                                    11 * hsh, 13 / 4, 3 * sh, 3 / 4],
                        atol=1e-15, rtol=0.0)
        assert_allclose(t_seg, [1 / 2, 1 / 12, -1 / 6, -1 / 8,
                                0, 9 / 8, 7 / 4, 9 / 10], atol=1e-15, rtol=0.0)
        assert_array_equal(t_seg, t_line)

        if plots:
            from fractions import Fraction

            def angle(a, b):
                return numpy.rad2deg(numpy.arctan(real_divide(*(b - a)[::-1],
                                                                   numpy.inf)))

            def annotate_projection(fig, ax, t, s, e, r, p):
                x, y = r
                hx = 0.15 + 0.1 * numpy.isclose(y, p[1])
                hy = 0.15 + 0.1 * numpy.isclose(x, p[0])
                x += hx * numpy.sign(x - p[0]) * ~numpy.isclose(x, p[0])
                y += hy * numpy.sign(y - p[1]) * ~numpy.isclose(y, p[1])
                if t:
                    f = Fraction(abs(t)).limit_denominator(10)
                    label = fr'\frac{{{f.numerator}}}{{{f.denominator}}}'
                else:
                    label = 0
                if t < 0:
                    label = '-' + label
                t = ax.text(x, y, f'$t = {label}$', ha='center', va='center',
                            rotation=angle(s, e))
                b = t.get_window_extent(fig.canvas.get_renderer()).\
                                        transformed(ax.transData.inverted())
                ax.update_datalim(b.corners())

            p = numpy.array(p)
            p1 = numpy.array(p1)
            p2 = numpy.array(p2)
            with plotting_context(constrained_layout=True) as fig:
                ax = fig.subplots(1)

                # Draw segments
                for t, s, e in zip(t_seg, p1, p2):
                    ax.axline(s, e, color='k', linestyle='--')
                    ax.plot([s[0], e[0]], [s[1], e[1]],
                            color='k', linestyle='-')

                    # Draw normals
                    r = self.point(t, s, e)
                    ax.plot([p[0], r[0]], [p[1], r[1]],
                            color='k', linestyle='--')

                    # Annontate projections
                    ax.plot(*r, color='k', marker='o',
                            markerfacecolor='w', markersize=10)
                    annotate_projection(fig, ax, t, s, e, r, p)

                # Draw points
                ax.plot(*p, color='k', linestyle='none',
                        marker='o', markerfacecolor='r', markersize=10)
                ax.plot(*p1.T, color='k', linestyle='none',
                        marker='o', markerfacecolor='gray', markersize=10)
                ax.plot(*p2.T, color='k', linestyle='none',
                        marker='o', markerfacecolor='gray', markersize=10)

                # Annotate point
                ax.text(*p + [-0.35, -0.15], fr'$\vec{{p}}$',
                        ha='center', va='center')

                ax.axis('equal')
                save(fig, __name__, 'n_lines')

    def test_broadcast(self):
        p = numpy.array([[0, 1], [1, 0], [0, -1], [-1, 0]])
        p1 = numpy.array([[-0.5, 0.5], [0.5, 0.5], [0.5, -0.5], [-0.5, -0.5]])
        p2 = numpy.array([[0.5, 0.5], [0.5, -0.5], [-0.5, -0.5], [-0.5, 0.5]])

        dist1 = segment_distance(p, p1, p2, axis=1)
        assert_allclose(dist1, [0.5, 0.5, 0.5, 0.5], atol=1e-15, rtol=0.0)

        hs2 = numpy.sqrt(1 / 2)

        dist2 = segment_distance(p[:, None, :], p1, p2, axis=-1)
        assert_allclose(dist2, [[0.5, hs2, 1.5, hs2],
                                [hs2, 0.5, hs2, 1.5],
                                [1.5, hs2, 0.5, hs2],
                                [hs2, 1.5, hs2, 0.5]], atol=1e-15, rtol=0.0)

    def test_error(self):
        p = [[1, 2], [3, 4], [5, 6]]
        p1 = [[1, 2], [3, 4]]
        p2 = [[1, 2], [3, 4]]
        with raises(ValueError, match='broadcast'):
            segment_distance(p, p1, p2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for working with new-style MS Excel documents, only available
when the ``[xlsx]`` :ref:`extra <installation-extras>` is installed.

If `openpyxl`_ is not found at import time, this module will have a
:py:data:`xlsx_enabled` attribute, which will be :py:obj:`False`. If
`openpyxl`_ is found, on the other hand, :py:data:`xlsx_enabled` will
be :py:obj:`True`, and all the dependent functions and attributes of the
module will be present.

.. py:data:: xlsx_enabled

   A boolean value indicating whether the ``[xlsx]``
   :ref:`extra <installation-extras>` has been installed. If
   :py:obj:`False`, the API will be severely limited.


.. include:: /link-defs.rst 
"""

__all__ = ['xlsx_enabled', 'EXTENSION', 'ensure_extension',]


from warnings import warn

from . import ensure_extension as _ensure_extension


try:
    from openpyxl.styles import Border
except ImportError:
    from .. import _display_missing_extra
    _display_missing_extra('xlsx', 'openpyxl')
    xlsx_enabled = False
else:
    __all__.extend(['apply_border', 'copy_range'])
    xlsx_enabled = True


#: The extension that will be appended to output file names by
#: :py:func:`ensure_extension`.
EXTENSION = '.xlsx'


def ensure_extension(output):
    """
    Verify that the output object is a valid file name, and return a
    fixed version if not.

    If `output` is a :py:class:`str` and does not end with ``'.xlsx'``,
    fix it. If it ends with ``'.xls'``, append ``'x'``, otherwise append
    ``'.xlsx'``. All other types are assumed to be proper file-like
    objects that are passed through.
    """
    if isinstance(output, str):
        return _ensure_extension(output, EXTENSION, partial_policy='append',
                                 partial_limit=1)
    return output


if xlsx_enabled:
    def copy_range(ws_in, row_in=None, column_in=None, width=None, height=None,
                   ws_out=None, row_out=None, column_out=None, delete_empty=True):
        """
        Copy a range of cells from one worksheet to another.

        All arguments besides `ws_in` are technically optional, with
        some restrictions. This allows shortcuts for copying blocks
        between or even within worksheets.

        .. warning::

           At this time, copying into the same worksheet may yield
           incorrect results if the source and destination overlap.

        Parameters
        ----------
        ws_in : openpyxl.worksheet.worksheet.Worksheet
            The source worksheet with the data.
        row_in : int or None
            The row of the upper-left hand corner in the source. If
            :py:obj:`None`, single-column mode will be used. In this
            case, `width` may not be specified and efectively becomes
            ``1``.
        column_in : int or None
            The column of the upper-left hand corner of the source. If
            :py:obj:`None`, single-row mode will be used. In this case,
            ``height`` may not be specified and efectively becomes
            ``1``.
        width : int or None
             The width (number of columns) of the range. If
             :py:obj:`None`, all available columns will be used.
        height : int or None
            The height (number of rows) of the range. If :py:obj:`None`,
            all available rows will be used.
        ws_out : openpyxl.worksheet.worksheet.Worksheet
            The destination worksheet. If omitted, defaults to the
            source worksheet. In that case, the location being copied to
            must be different.
        row_out : int or None
            The row of the upper-left hand corner in the destination. If
            omitted, defaults to the same location as the source.
        column_out : int or None
            The column of the upper-left hand corner in the destination.
            If omitted, defaults to the same location as the source.
        delete_empty : bool
            Whether or not to remove destination cells that are
            overwritten by missing cells in the source range. Default is
            :py:obj:`True`.


        Only one of `row_in` and `column_in` may be :py:obj:`None`. All
        indices are one-based to conform to openpyxl notation.

        Return
        ------
        n : int
            The number of non-empty source cells copied. This will be zero
            if the destination is the same as the source.
        """
        if row_in is None and column_in is None:
            raise ValueError('Either `row_in` or `col_in` must be specified')

        if row_in is None:
            # Column-only mode
            if width is not None:
                raise ValueError('`width` not allowed for column copy')
            row_in = 1
            mr, mc = None, column_in
        elif column_in is None:
            # Row-only mode
            if height is not None:
                raise ValueError('`height` not allowed for row copy')
            column_in = 1
            mr, mc = row_in, None
        else:
            mr = None if height is None else row_in + height - 1
            mc = None if width is None else column_in + width - 1

        if row_out is None:
            row_out = row_in
        if column_out is None:
            column_out = column_in

        if ws_in == ws_out and row_in == row_out and column_in == column_out:
            warn('Source and destination are the same. Skipping.')
            return 0

        iterator = ws_in.iter_rows(min_row=row_in, max_row=mr,
                                   min_col=column_in, max_col=mc)

        counter = 0
        for dest_r, row in enumerate(iterator, start=row_out):
            for dest_c, item in enumerate(row, start=column_out):
                value = item.value
                copy = value is not None
                if copy or delete_empty:
                    ws_out.cell(row=dest_r, column=dest_c).value = value
                    if copy:
                        counter += 1

        return counter


    def apply_border(ws, start_row, end_row, start_column, end_column, *,
                     merge=False, **kwargs):
        """
        Apply a border around the specified range of cells as if it was
        a single object.

        The border may be specified in a number of ways through keyword
        arguments.

        If the argument `border` is specified, there may not be any
        other `kwargs`. If `border` is a
        :py:class:`openpyxl.styles.borders.Side`, apply it on all sides.
        If it is a :py:class:`openpyxl.styles.borders.Border`, apply the
        attributes of the border to the sides they correspond to around
        the whole edge.

        The other option is to specify `kwargs` as some subset of
        {`top`, `left`, `bottom`, `right`}. Each argument can be a
        :py:class:`openpyxl.styles.borders.Side` or an
        :py:class:`openpyxl.styles.borders.Border`. Only the part of the
        border corresponding to the side of the object it goes on will
        be used. For example, only the
        :py:attr:`~openpyxl.styles.borders.Border.top` attribute of a
        border specified for `top` will be used. The other attributes
        will be ignored. The borders corresponding to missing or
        :py:obj:`None` arguments will not be changed in this case.

        This function is based heavily on the recipe described in

        http://openpyxl.readthedocs.io/en/stable/styles.html#styling-merged-cells.
        """

        def check_side(side, direction):
            """
            If passed a `Side`, convert into a border with only that direction set.
            If passed a `Border`, effectively unset all the other directions.
            If `None`, pass-through.
            """
            if side is None:
                return side
            if isinstance(side, Border):
                side = getattr(side, direction)
            return Border(**{direction: side})

        if len(kwargs) == 1 and 'border' in kwargs:
            top = left = bottom = right = kwargs['border']
        elif set(kwargs) <= {'top', 'left', 'bottom', 'right'}:
            top = kwargs.get('top')
            left = kwargs.get('left')
            bottom = kwargs.get('bottom')
            right = kwargs.get('right')
        else:
            raise ValueError('Found illegal keyword arguments.')

        if merge:
            # At least in LibreOffice, the border does not come out right if it
            # is applied before the merge, so merge first, apply border after.
            ws.merge_cells(start_row=start_row, end_row=end_row,
                           start_column=start_column, end_column=end_column)

        top = check_side(top, 'top')
        left = check_side(left, 'left')
        bottom = check_side(bottom, 'bottom')
        right = check_side(right, 'right')

        for row in range(start_row, end_row + 1):
            if left:
                cell = ws.cell(row=row, column=start_column)
                cell.border = cell.border + left
            if right:
                cell = ws.cell(row=row, column=end_column)
                cell.border = cell.border + right

        for column in range(start_column, end_column + 1):
            if top:
                cell = ws.cell(row=start_row, column=column)
                cell.border = cell.border + top
            if bottom:
                cell = ws.cell(row=end_row, column=column)
                cell.border = cell.border + bottom



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 09 Jan 2021: Initial Coding

"""
Recipes for handling different types of configuration files.
"""


__all__ = [
    'JSONConfiguration', 'JSONObject', 'NumpyObject', 'json_registry'
]


from collections.abc import Mapping
from io import StringIO
from json import load
from numbers import Number
from os import replace, sep
from os.path import isfile
from sys import maxsize, stdout

import numpy as np

from .files import open_file
from .mapping import Namespace
from .recipes import islast
from .structures import Trie


class JSONObject:
    """
    Base class of additional output formatting types.

    To register a new formatter use the module-level
    :py:meth:`json_registry.register` function.

    .. py:attribute:: type

       The type or types supported by this formatter. Any object that is
       a valid second argument to :py:func:`isinstance` is accepted.
    """
    __slots__ = ('type')

    def __init__(self, type):
        """
        Create a formatting object for the specified type.

        Parameters
        ----------
        type : type or tuple
            The permitted type or types. Any object that is a valid
            second argument to :py:func:`isinstance` is accepted.
        """
        self.type = type

    def format(self, file, obj, prefix, indent):
        """
        Format an object for output to a file.

        The base class method is a no-op and should be overriden by
        subclasses.

        Parameters
        ----------
        file : file-like
            An open file-like object to write to.
        obj :
            An instance of :py:attr:`type` to format.
        prefix : str
            The line prefix (current indentation) to start with.
        indent : int
            The number of spaces to use for any additional indents.
        """
        pass


class NumpyObject(JSONObject):
    """
    Type of :py:class:`JSONObject` speficic to numpy arrays.

    An instance of this class is registered automatically.
    """

    #: Format spec to use for numpy arrays.
    #:
    #: Must not contain a ``'prefix'`` key, which will be silently
    #: purged if present.
    _PRINTOPTS = {
        'separator': ', ',
        'floatmode': 'fixed',
        'precision': 18,
        'suppress_small': False,
        'threshold': maxsize,
        'sign': ' ',
        'max_line_width': 120,
        'formatter': {
            'bool': lambda x: str(x).lower()
        }
    }

    def __init__(self, printopts=None, type=np.ndarray):
        """
        Create an instance with the specified print options.

        Parameters
        ----------
        printopts : dict, optional
            A formatting spec that contains the keyword arguments to
            :py:func:`numpy.array2string`. All keys listed are
            allowed, except ``'prefix'``. The default is to use the
            builtin formatting spec.
        """
        super().__init__(type=type)
        if printopts:
            self.printopts = self._PRINTOPTS.copy()
            self.printopts.update(printopts)
        else:
            self.printopts = self._PRINTOPTS
        self.printopts.pop('prefix', None)

    def format(self, file, obj, prefix, indent):
        """
        Pretty-print a numpy array with the specifed indentation.
        """
        prefix += ' ' * indent
        file.write('\n')
        file.write(prefix)
        file.write(np.array2string(obj, prefix=prefix, **self.printopts))


class _JSONRegistry(list):
    def register(self, obj):
        self.append(obj)


#: A list-like object with an additional
#: :py:meth:`~json_registry.register` method.
#:
#: .. py:method:: json_registry.register(formatter: JSONObject)
#:
#:    Call this function to add output types to be used with
#:    :py:meth:`JSONConfiguration._pprint`.
#:
#: .. py:method:: json_registry.clear()
#:
#:    Call this function to clear the registry.
#:
#: An instance of :py:class:`NumpyObject` with default parameters is
#: registered automatically.
json_registry = _JSONRegistry()

del _JSONRegistry


json_registry.register(NumpyObject())


def _make_trie(iterable):
    """
    Add all the items in `iterable` into a
    :py:class:`~haggis.structures.Trie`. Hashable elements are treated
    as single top-level leaves. Lists are treated as nested items.
    """
    t = Trie()
    for item in iterable:
        if not isinstance(item, list):
            item = [item]
        t.add(item)
    return t


class JSONConfiguration(Namespace):
    """
    Class for managing loading and updating JSON configurations.

    The file into its just loaded into the namespace `__dict__`. The
    metadata attributes are stored in `__slots__`, so do not interfere
    with the configuration keys. In particular, the following metadata
    key is supported:

    .. py:attribute:: _source

       The source file or mapping for the configuration.

    Child classes should extend :py:meth:`_reload` to parse the necessary
    keys directly in their own namespace, and to do any error checking
    on the loaded data. Conversely, children should extend
    :py:meth:`_update` to perform any additional steps necessary to
    serialize back into a file.

    Methods and attibutes are not private: the single underscores are
    intended to reduce the probablility of shadowing a configuration
    key.
    """
    __slots__ = ('_source',)

    def __init__(self, source):
        """
        Initialize a new instance with the specified source.

        Parameters
        ----------
        source : str or Path or Mapping
            If a mapping type, use to update the instance
            dictionary directly. Otherwise assume filename or path.
        """
        self._reload(source)

    def __repr__(self):
        """
        Return a programmatic representation of this configuration.

        Returns
        -------
        str
            A string representation of the object.
        """
        if isinstance(self._source, str) and sep in self._source:
            f = "r'{}'".format(self._source)
        else:
            f = repr(self._source)
        return '{}({})'.format(type(self).__name__, f)

    def __str__(self):
        """
        A detailed, expanded, representation of this configuration.

        Returns
        -------
        str :
            A string containing the contents of this mapping.
        """
        s = StringIO()
        self._pprint(s)
        return s.getvalue()

    def _reload(self, source=None):
        """
        Load the specified file (or the default :py:attr:`_filename`).

        Any nested mapping objects are replaced with
        :py:class:`~haggis.mapping.Namespace`.

        Override this method to implement additional conversion or
        error checking functionality.

        Parameters
        ----------
        source : str or Path or Mapping, optional
            If supplied, replace the default source for this
            configuration.
        """
        def emplace(mapping):
            """ Convert all nested dicts to Namespaces """
            for k, v in mapping.items():
                if isinstance(v, Mapping):
                    v = Namespace(emplace(v))
                yield k, v

        if source:
            self._source = source
        self.__dict__.clear()
        if isinstance(self._source, Mapping):
            data = self._source
        else:
            with open(self._source, 'r') as f:
                data = load(f)

        self.__dict__.update(emplace(data))

    def _update(self, source=None, *exclude):
        """
        Write the dictionary back to the file or original mapping.

        If writing to a mapping, convert nested
        :py:class:`~haggis.mapping.Namespace`\\ s to :py:class:`dict`.

        Override this method to implement additional conversion or
        error checking functionality.

        Parameters
        ----------
        source : str or Path or Mapping, optional
            If provided, supplies a non-default destination for the
            namespace, but does not permanently replace
            :py:attr:`_source`. The default is `None`.
        *exclude :
            sequence of keys to exclude. Hashable types are keys in the
            current dictionary. Lists indicate multi-level keys. For
            example, to avoid printing ``self.a.b``, add an exclude
            ``['a', 'b']``.
        """
        if not source:
            source = self._source

        if isinstance(source, Mapping):
            exclude = _make_trie(exclude)
            def unplace(prefix, d):
                for k, v in d.items():
                    item = prefix + [k]
                    if item in exclude:
                        continue
                    if isinstance(v, Namespace):
                        v = dict(unplace(item, v.__dict__))
                    yield k, v
            source.clear()
            source.update(unplace([], self.__dict__))
        else:
            if isfile(source):
                end = b'.bak' if isinstance(source, (bytes, bytearray)) else '.bak'
                replace(source, source + end)
            self._pprint(source, exclude=exclude)

    def _check_path(self, *keys):
        """
        Ensure that a given sequence of nested namespaces exists in
        this namespace.

        All keys in the sequence besides the last one must contain a
        mutable :py:attr:`~object.__dict__`. If the last item in the
        sequence does not exist, it will be created as a
        :py:class:`~haggis.mapping.Namespace`.

        Parameters
        ----------
        *keys :
            Sequence of keys to verify. Any missing keys will be
            created as empty namespaces. An error will be raised if
            an intermediate object exists without a mutable `__dict__`.

        Return
        ------
        The object at the end of the chain. If `keys` is empty, returns
        the current object.
        """
        obj = self
        for key in keys:
            if key not in obj.__dict__:
                obj.__dict__[key] = Namespace()
            obj = obj.__dict__[key]
        return obj

    def _pprint(self, filename=None, *, indent=4, root_indent=False, linewidth=120,
                float_format='', int_format='', bool_format=True,
                bytes_format='utf-8', exclude=()):
        """
        Pretty print the configuration into a file.

        All arguments besides the file name are keyword-only. Output
        formats for additional datatypes besides the normal JSON types
        are supported by registering :py:class:`JSONObject` descriptors
        using :py:meth:`json_registry.register`.

        Parameters
        ----------
        filename : str or Path or file-like
            The file to write to. If an open file, must have write
            permissions. If a string or path, will be truncated or
            created (using ``'w'`` mode).
        indent : int
            The number of spaces to indent nested objects by. Default 4.
        root_indent : bool
            Whether or not to indent the root namespace. Default is not
            to.
        linewidth : int
            The number of characters to attempt to wrap arrays at.
            Default is 120.
        bytes_format : str
            Name of the encoding to use to convert byte arrays to
            string. If `None`, record byte arrays as arrays of
            hexadecimal integers.
        exclude : iterable
            An iterable of items to exclude from the printout. Hashable
            elements are interpreted as top-level keys. Nested elements
            must be specified as a list of keys.
        """
        def iscontainer(e):
            return isinstance(e, (dict, list, tuple, np.ndarray))

        def strlen(e):
            if isinstance(e, str):
                return len(e)
            if isinstance(e, (bytes, bytearray)):
                if bytes_format is None:
                    return 4 + 6 * len(e)
                return len(e.decode(bytes_format))
            return 0

        def pprint_object(obj, key, spaces):
            """
            Spaces is the indented number of spaces!

            Why: this allows the root element to have zero indentation while
            all the other elements have correct indentation.
            """
            f.write('{\n')
            if key is None:
                iterator = obj.items()
            else:
                iterator = ((k, v) for k, v in obj.items() \
                                            if key + [k] not in exclude)
            for last, (k, v) in islast(iterator):
                f.write(spaces)
                f.write('"' + str(k) + '":')
                f.write(' ')

                pprint_element(v, None if key is None else key + [k], spaces)
                if not last:
                    f.write(',\n')

            spaces = '' if len(spaces) <= indent else spaces[:-indent]
            f.write('\n' + spaces + '}')

        def pprint_array(arr, spaces):
            more_spaces = spaces + ' ' * indent
            f.write('[')

            multiline = (
                any(map(iscontainer, arr)) or
                sum(map(strlen, arr)) + 2 * len(arr) +
                                        len(more_spaces) > linewidth
            )

            if multiline:
                prefix = more_spaces
                suffix = '\n'
            else:
                prefix = ''
                suffix = ' '
                spaces = ''
            comma = ',' + suffix

            f.write(suffix)
            for last, e in islast(arr):
                f.write(prefix)
                pprint_element(e, None, more_spaces)
                if not last:
                    f.write(comma)
            f.write(suffix + spaces + ']')

        def pprint_bytes(arr, spaces):
            more_spaces = spaces + ' ' * indent
            f.write('[')

            if 4 + 6 * len(arr) + len(more_spaces) > linewidth:
                prefix = more_spaces
                suffix = '\n'
            else:
                prefix = ''
                suffix = ' '
                spaces = ''
            comma = ',' + suffix

            f.write(suffix)
            for last, b in arr:
                f.write(prefix + '{:%0.2X}'.format(b))
                if not last:
                    f.write(comma)
            f.write(suffix + spaces + ']')

        def pprint_element(e, key, spaces):
            if e is None:
                f.write('null')
            elif isinstance(e, bool):
                if bool_format:
                    f.write(str(e).lower())
                else:
                    f.write(str(int(e)))
            elif isinstance(e, (bytes, bytearray)):
                if bytes_format is None:
                    pprint_bytes(e, spaces)
                else:
                    f.write('"' + esc(e.decode(bytes_format)) + '"')
            elif isinstance(e, str):
                f.write('"' + esc(e) + '"')
            elif isinstance(e, (int, np.integer)):
                f.write('{:{}}'.format(e, int_format))
            elif isinstance(e, Number):
                f.write('{:{}}'.format(e, float_format))
            elif isinstance(e, (list, tuple)):
                pprint_array(e, spaces)
            elif isinstance(e, (dict, Namespace)):
                pprint_object(e, key, spaces + ' ' * indent)
            else:
                for reg in json_registry:
                    if isinstance(e, reg.type):
                        reg.format(f, e, spaces, indent)
                        break
                else:
                    raise TypeError(
                        "Don't know how to encode {}".format(type(e).__name__)
                    )

        def esc(s):
            return s.replace('\\', '\\\\')

        exclude = _make_trie(exclude)
        with open_file(filename or stdout, 'w') as f:
            pprint_object(self, [], ' ' * indent * root_indent)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for extending and configuring the Python XML framework.
"""

__all = ['SAXBase', 'SAXLoggable']


from xml.sax.handler import ContentHandler

from ..logs import MetaLoggableType


class SAXBase(ContentHandler):
    """
    An empty SAX parser with some convenience functionality built in.

    This class provides a reference to the locator. All of the actual SAX
    callback methods are currently no-ops.
    """
    def __init__(self):
        """
        Initialize the base classes and set the locator to `None`.
        """
        super().__init__()
        self.locator = None

    def setDocumentLocator(self, locator):
        """
        Set the locator used in error reporting.

        The locator can be accessed through the :py:meth:`locate`
        message-reformatting utility method.
        """
        self.locator = locator

    def _get_position(self):
        """
        Retrieve the position based on the currently configured locator.

        If there is no locator set, the position will be NaN.
        """
        if self.locator is None:
            return (float('nan'),) * 2
        return self.locator.getLineNumber(), self.locator.getColumnNumber()

    def locate(self, message, *args, short=False):
        """
        Convert a message and argument-list to a message with location
        information, and extends the argument list appropriately.

        Return a tuple containing two elements: the message and the
        argument list as a single tuple that can be expanded into the
        argument of any of the logging methods.

        The full version of the location string (default) includes the
        file name, the line number and the column. The short version
        only inculdes the line number and column.
        """
        if short:
            message += ' on line %d:%d'
            extra = self._get_position()
        else:
            message += ' in %s:%d:%d'
            extra = (self.locator.getSystemId(), *self._get_position())
        return (message, (*args, *extra))

    def short_locate(self, message, *args):
        """
        Identical to :py:meth:`locate`, except that the file name is not
        included in the updated message.
        """
        return (message + ' on line %d:%d', (*args, *self._get_position()))


class SAXLoggable(SAXBase, metaclass=MetaLoggableType):
    """
    A type of :py:class:`SAXBase` that provides logging in addition to
    location methods.
    """
    def log(self, level, msg, *args, **kwargs):
        """
        Append location information to a log message.

        This method allows an additional keyword argument `short` that
        determines whether or not the full file name will be present in
        the location. `short` is :py:obj:`True` by default, meaning that
        only the line and column number are reported.
        """
        short = kwargs.pop('short', 'True')
        msg, args = self.locate(msg, *args, short=short)
        self.logger.log(level, msg, *args, **kwargs)

    def setDocumentLocator(self, locator, level=None):
        """
        Set the locator used in error reporting and log it, with
        location.

        Logging is done using :py:meth:`log`. The locator will not be
        logged if `level` is :py:obj:`None`.
        """
        super().setDocumentLocator(locator)
        if level is not None:
            self.log(level, "Set document locator")



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 20 Jan 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.objects` module.
"""


from array import array
from itertools import islice, product
from string import ascii_lowercase
import sys

import numpy

from ..objects import getsizeof


class TestGetsizeof:
    def setup_method(self):
        """
        Find the size of 'a', 'b', 1, 2 combined.
        """
        self.size = (
             sys.getsizeof('a') + sys.getsizeof(1) +
             sys.getsizeof('b') + sys.getsizeof(2)
        )

    def test_simple(self):
        """
        Show that simple objects have system size.
        """
        for obj in 32, None:
            assert getsizeof(obj) == sys.getsizeof(obj)

    def test_mapping(self):
        """
        Show that mappings are the sum of their components.
        """
        obj = {'a': 1, 'b': 2}
        assert getsizeof(obj) == sys.getsizeof(obj) + self.size

    def test_iterable(self):
        """
        Show that iterables are the sum of their components.
        """
        base = {'a', 1, 'b', 2}
        for c in list, set, frozenset, tuple:
            obj = c(base)
            assert getsizeof(obj) == sys.getsizeof(obj) + self.size

    def test_array(self):
        """
        Show that all array/string types have system size.
        """
        obj = b'test'
        assert getsizeof(obj) == sys.getsizeof(obj)
        obj = bytearray(obj)
        assert getsizeof(obj) == sys.getsizeof(obj)
        obj = array('b', obj)
        assert getsizeof(obj) == sys.getsizeof(obj)
        obj = str(obj)
        assert getsizeof(obj) == sys.getsizeof(obj)
        obj = numpy.array(list(obj))
        assert getsizeof(obj) == sys.getsizeof(obj)

    def test_loop(self):
        """
        Show that loops are handled correctly.
        """
        obj = {'a': 1, 'b': 2, 'c': None}
        obj['c'] = obj
        assert getsizeof(obj) == sys.getsizeof(obj) + self.size + \
                                 sys.getsizeof('c')

    def test_duplicate(self):
        """
        Show that duplicates are handled correctly.
        """
        obj = 10 * ['a', 1, 'b', 2]
        assert getsizeof(obj) == sys.getsizeof(obj) + self.size

    def test_nested(self):
        """
        Shows mixed types with nesting.
        """
        obj = {'a': 1, 'b': [], 'c': ({'d', 'e'}, {'f': 6})}
        size = (sys.getsizeof(obj) +
            sys.getsizeof('a') + sys.getsizeof(1) +
            sys.getsizeof('b') + sys.getsizeof(obj['b']) +
            sys.getsizeof('c') + sys.getsizeof(obj['c']) +
                                 sys.getsizeof(obj['c'][0]) +
                                     sys.getsizeof('d') + sys.getsizeof('e') +
                                 sys.getsizeof(obj['c'][1]) +
                                     sys.getsizeof('f') + sys.getsizeof(6)
        )
        assert getsizeof(obj) == size

    def test_handlers(self):
        """
        Verify that handler are more specific than registered classes.
        """
        def tup_handler(t):
            yield from (x * 3 for x in t)

        obj = [((), 33), ('b', ['c'])]
        # Warning: sys.getsizeof(3 * ['c']) < sys.getsizeof(['c', 'c', 'c'])
        size = (
            sys.getsizeof(obj) + 2 * sys.getsizeof((None, None)) +
            sys.getsizeof(()) + sys.getsizeof(99) +
            sys.getsizeof('bbb') +
                sys.getsizeof(['c'] * 3) + sys.getsizeof('c')
        )
        assert getsizeof(obj, handlers=[(tuple, tup_handler)]) == size

    from .. import objects
    if hasattr(objects, 'ndarray_handler'):
        def test_ndarray(self):
            dtype1 = numpy.dtype([('a', '<f4'), ('b', 'O', 3), ('c', '<i4')])
            dtype2 = numpy.dtype([('d', bool, 3), ('e', dtype1, (2, 2)), ('f', 'O')])

            obj = numpy.empty((2, 2), dtype=dtype2)

            n = 2 * 2 * 2 * 2 * 3
            # Ensure each string is different regardless of interning
            obj['e']['b'] = numpy.array(
                    list(''.join(e) for e in islice(
                            product(ascii_lowercase, repeat=2), None, n
                    ))
            ).reshape(2, 2, 2, 2, 3)
            # Ensure some references are copied 2 times
            obj['f'] = ['a', 'b']

            size = (sys.getsizeof(obj) +
                    n * sys.getsizeof('zz') +
                    2 * sys.getsizeof('z'))
            assert getsizeof(obj) == size



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 09 Jan 2021: Initial Coding

"""
Utilities and recipes for extending ctypes functionality.
"""


__all__ = [
    'c_bool_p', 'c_byte_p', 'c_double_p', 'c_float_p', 'c_int_p',
    'c_int8_p', 'c_int16_p', 'c_int32_p', 'c_int64_p', 'c_long_p',
    'c_longdouble_p', 'c_longlong_p', 'c_short_p', 'c_size_t_p',
    'c_ssize_t_p', 'c_ubyte_p', 'c_uint_p', 'c_uint8_p', 'c_uint16_p',
    'c_uint32_p', 'c_uint64_p', 'c_ulong_p', 'c_ulonglong_p',
    'c_ushort_p',
    'check_zero', 'check_nonzero', 'decode_c_char_p', 'make_enum',
    'to_c_char_p', 'CDLLWrapper',
]


from contextlib import ExitStack
from ctypes import (
    CDLL, POINTER,
    c_int8, c_int16, c_int32, c_int64,
    c_byte, c_short, c_int, c_long, c_longlong,
    c_uint8, c_uint16, c_uint32, c_uint64,
    c_ubyte, c_ushort, c_uint, c_ulong, c_ulonglong,
    c_bool, c_size_t, c_ssize_t,
    c_float, c_double, c_longdouble,
    c_void_p,
)
from os.path import join
from os import name as os_name

from .recipes import CloseableMixin


c_byte_p = POINTER(c_byte)
c_short_p = POINTER(c_short)
c_int_p = POINTER(c_int)
c_long_p = POINTER(c_long)
c_longlong_p = POINTER(c_longlong)

c_int8_p = POINTER(c_int8)
c_int16_p = POINTER(c_int16)
c_int32_p = POINTER(c_int32)
c_int64_p = POINTER(c_int64)

c_ubyte_p = POINTER(c_ubyte)
c_ushort_p = POINTER(c_ushort)
c_uint_p = POINTER(c_uint)
c_ulong_p = POINTER(c_ulong)
c_ulonglong_p = POINTER(c_ulonglong)

c_uint8_p = POINTER(c_uint8)
c_uint16_p = POINTER(c_uint16)
c_uint32_p = POINTER(c_uint32)
c_uint64_p = POINTER(c_uint64)

c_bool_p = POINTER(c_bool)
c_size_t_p = POINTER(c_size_t)
c_ssize_t_p = POINTER(c_ssize_t)
c_float_p = POINTER(c_float)
c_double_p = POINTER(c_double)
c_longdouble_p = POINTER(c_longdouble)


def check_zero(fail_msg, error=ValueError, name_prefix=True):
    """
    Generate an error checker for the specified message and error type.

    The resulting function can be set as the
    :py:attr:`~ctypes._FuncPtr.errcheck` of a :py:mod:`ctypes` function.
    It will raise an error on truthy return values and pass through
    zeros.

    Parameters
    ----------
    fail_msg : str
        Messages may be new-style interpolation strings that index the
        function arguments and include the names ``__func__`` and
        ``__value__`` as a keywords.
    error : type
        The type of error to raise if the result is truthy.
    name_prefix : str or bool
        If `name_prefix` is a string, it gets prepended to the message
        directly. If any other truthy value, ``'{__func__}: '`` gets
        prefixed instead. Falsy values don't modify the message at all.
        String prefixes can contain interpolations themselves.

    Return
    ------
    callable :
        A function named `check_zero` that accepts arguments named
        ``value``, ``func``, and ``arguments`` and raises an error if
        ``value`` is truthy.
    """
    if isinstance(name_prefix, str):
        fail_msg = name_prefix + fail_msg
    elif name_prefix:
        fail_msg = '{__func__}: ' + fail_msg

    def check(value, func, arguments):
        if value:
            raise error(fail_msg.format(
                    *arguments, __func__=func.__name__, __value__=value
                ))
        return value

    check.__name__ = check.__qualname__ = 'check_zero'
    return check


def check_nonzero(fail_msg, error=ValueError, name_prefix=True):
    """
    Generate an error checker for the specified message and error type.

    The resulting function can be set as the
    :py:attr:`~ctypes._FuncPtr.errcheck` of a :py:mod:`ctypes` function.
    It will raise an error on falsy return values and pass through
    the return value otherwise.

    Parameters
    ----------
    fail_msg : str
        Messages may be new-style interpolation strings that index the
        function arguments and include the name `__func__` as a keyword.
    error : type
        The type of error to raise if the result is falsy.
    name_prefix : str or bool
        If `name_prefix` is a string, it gets prepended to the message
        directly. If any other truthy value, ``'{__func__}: '`` gets
        prefixed instead. Falsy values don't modify the message at all.
        String prefixes can contain interpolations themselves.

    Return
    ------
    callable :
        A function named `check_nonzero` that accepts arguments named
        ``value``, ``func``, and ``arguments`` and raises an error if
        ``value`` is falsy.
    """
    if isinstance(name_prefix, str):
        fail_msg = name_prefix + fail_msg
    elif name_prefix:
        fail_msg = '{__func__}: ' + fail_msg

    def check(value, func, arguments):
        if not value:
            raise error(fail_msg.format(*arguments, __func__=func.__name__))
        return value

    check.__name__ = check.__qualname__ = 'check_nonzero'
    return check


def make_enum(enum_type):
    """
    Generate an error checker that converts to the specified enum.

    The resulting function can be set as the
    :py:attr:`~ctypes._FuncPtr.errcheck` of a :py:mod:`ctypes` function.
    It converts the return value into a Python enum.

    Checkers returned by this function are cached, for efficiency when
    using multiple times on the same `enum_type`.

    Parameters
    ----------
    enum_type : callable
        Normally, this is a subclass of :py:class:`IntEnum`, which
        converts a C return value into the appropriate type. However,
        this may be any arbirary callable, as long as it has a
        ``__name__`` attribute.

    Return
    ------
    callable :
        A function named `make_` followed by the ``__name__`` of
        `enum_type` that accepts arguments named ``value``, ``func``,
        and ``arguments`` and converts ``value`` to the target type.
    """
    check = make_enum.registry.get(enum_type)
    if check is None:
        def check(value, func, arguments):
            return enum_type(value)
        make_enum.registry[enum_type] = check
        check.__name__ = check.__qualname__ = 'make_' + enum_type.__name__
    return check

make_enum.registry = {}


def decode_c_char_p(encoding='utf-8', null_error=False):
    """
    Generate an error checker that decodes strings with a custom
    encoding.

    The resulting function can be set as the
    :py:attr:`~ctypes._FuncPtr.errcheck` of a :py:mod:`ctypes` function.
    It decodes the bytes of a `char *` to a python string.

    Checkers returned by this function are cached, for efficiency when
    using multiple times on the same encoding.

    Parameters
    ----------
    encoding : str
        The name of the target encoding.
    null_error : bool
        Whether or not to raise an error on NULL strings. Defaults is
        `False`.

    Return
    ------
    callable :
        A function named `decode_as_` followed by the target encoding
        that accepts arguments named ``value``, ``func``, and
        ``arguments`` and decodes ``value`` using the specified
        encoding.
    """
    key = encoding.casefold().replace('-', '_')
    if null_error:
        key += '_err'
    decoder = decode_c_char_p.registry.get(key)
    if decoder is None:
        if null_error:
            def decode(value, func, arguments):
                if value is None:
                    raise TypeError(f'{func.__name__}: expected string result, but got None')
                return value.decode(encoding)
        else:
            decode = lambda value, func, arguments: value.decode(encoding)
        decode.__name__ = decode.__qualname__ = f'decode_as_{key}'
        decoder = decode_c_char_p.registry[key] = decode
    return decoder

decode_c_char_p.registry = {}


def to_c_char_p(s, encoding='utf-8'):
    """
    Convert a string into a NUL-terminated :py:class:`bytes`.

    Parameters
    ----------
    s : str
        The string to convert.
    encoding : str
        The encoding to use. Default is ``utf-8``.

    Returns
    -------
    bytes
        The encoded string, terminated by ``b'\x00'``.
    """
    if s is None:
        return None
    return s.encode(encoding) + b'\x00'


if os_name == 'nt':
    from ctypes import windll, wintypes
    from os import add_dll_directory as _add_dll_directory
    _close_dll_handle = windll.kernel32.FreeLibrary
    _close_dll_handle.argtypes = [wintypes.HMODULE]
    _close_dll_handle.restype = wintypes.BOOL
    _close_dll_handle.errcheck = check_nonzero('Unable to release module',
                                               WindowsError)
else:
    def add_dll_directory(*args, **kwargs):
        return CloseableMixin()
    _close_dll_handle = CDLL('').dlclose
    _close_dll_handle.argtypes = [c_void_p]
    _close_dll_handle.restype = c_int
    _close_dll_handle.errcheck = check_zero(
        'Unable to release module: error {__value__}', OSError
    )


class CDLLWrapper(CDLL, CloseableMixin):
    """
    Wrapper for full-path DLLs to make them closeable and manage
    dependencies.

    This class is mostly useful for Windows, since it manages
    dependency folders. The only purpose it serves on UNIX-like
    systems is to close the library handle.
    """

    def __init__(self, name, *folders, prefix='', index=None):
        """
        Create a wrapper for the specified library.

        Parameters
        ----------
        name : str
            The basename of the library. Extension is optional.
        folders : str
            Directories containing the dependencies of the main library.
            Ignored on non-Windows systems.
        prefix : str
            The folder relative to which all `folders` should be
            resolved. Default is empty.
        index : int
            The element of `folders` containing the library. If `None`,
            the library name does not require or already has a folder
            prefixed to it.
        """
        if prefix:
            self._folders = tuple(join(prefix, folder) for folder in folders)
        else:
            self._folders = tuple(folders)
        self._basename = name
        self._index = index

        self.open()

    def open(self):
        """
        (Re)open this library, if it is closed.

        If already open, this is a no-op.

        On Windows, the dependency folders are added prior to opening
        the DLL itself.
        """
        if not self._handle:
            lib = self._basename if self._index is None \
                        else join(self._folders[self._index], self._basename)
            with ExitStack() as dependencies:
                for folder in self._folders:
                    dependencies.enter_context(_add_dll_directory(folder))
                CDLL.__init__(self, lib)  # Always open full name
                self._dependencies = dependencies.pop_all()

    def close(self):
        """
        Close this library, if it is open.

        If not open, this is a no-op.

        Any cached function objects are deleted, and the underlying
        binary is closed.

        On Windows, the dependency folders are unloaded after closing
        the DLL itself.
        """
        if self._handle:
            _close_dll_handle(self._handle)
            self._handle = 0
            self._dependencies.close()
    
            # Clear cached objects
            for name, value in list(self.__dict__.items()):
                if isinstance(value, self._FuncPtr):
                    delattr(self, name)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Routines for dealing with file types.

The module names in this package generally correspond to the informal
name of the file type they deal with, or to the extension.
"""

__all__ = [
    'ensure_extension', 'insert_suffix', 'open_file', 'split_extension',
    'PreOpenedFile',
]


from os.path import splitext

from ..recipes import is_ordered_subset as _is_ordered_subset
from ..string_util import check_value


def ensure_extension(name, ext, partial_policy=None, partial_limit=None):
    """
    Verify that the name ends with the required extension, and
    update it if not.

    `name` is assumed to be a string. `ext` is the desired suffix,
    usually beginning with ``'.'``. If `name` aleady ends with `ext`,
    return it as-is. If not, the extension will either be appended or
    completed depending on `partial_policy` and `partial_limit`. The
    first character of `ext` is treated as the separator character. It
    must appear in `name` for any of the completions to work.

    Recognized values for `partial_policy` are as follows (case
    insensitive):

        :py:obj:`None`, ``'none'``, ``''``
            No partial extensions are recognized. If ``name='a.xls'``
            and ``ext='.xlsx'``, the result is ``'a.xls.xlsx'``.
            Similarly, for ``'b.jpg'``, ``'.jpeg'``, the result is
            ``'b.jpg.jpeg'``.
        ``'append'``, ``'+'``
            Existing extension can be extended by at most
            `partial_limit` characters to acheive the target. If
            ``name='a.xls'`` and ``ext='.xlsx'``, the result is
            ``'a.xlsx'``. However, for ``'b.jpg'``, ``'.jpeg'``, the
            result is ``'b.jpg.jpeg'``.
        ``'insert'``, ``'^'``
            Existing extension can have up to `partial_limit` characters
            inserted anywhere to achieve the target. If ``name='a.xls'``
            and ``ext='.xlsx'``, the result is ``'a.xlsx'``. Similarly,
            for ``'b.jpg'``, ``'.jpeg'``, the result is ``b.jpeg``.
        ``'strip'``, ``'-'``
            The existing extension may be loner than the desired one, so
            up to `partial_limit` characters may be stripped off the end
            to match the target. If ``name='a.xlsx'`` and
            ``ext='.xls'``, the result is ``'a.xls'``. However, for
            ``'b.jpeg'``, ``'.jpg'``, the result is ``'b.jpeg.jpg'``.
        ``'remove'``, ``'x'``
            The existing extension may be longer than the desired one,
            so removing up to `partial_limit` characters anywhere in the
            name is allowed. If ``name='b.jpeg'`` and ``ext='.jpg'``,
            the result is ``'b.jpg'``. However, for ``name='b.jpg'`` and
            ``ext='.jpeg'``, the result is ``'b.jpg.jpeg'``.
        ``'replace'``, ``'r'``
            Replace any existing extension with the provided one.
        ``'create'``, ``'c'``
            Create the extension only if one does not already exist.

    `partial_limit` determines the maximum number of characters that can
    be modified to achieve the target. If :py:obj:`None` or a number
    greater than the length of `ext` mean "any number". If zero, the
    result is the same as for `partial_policy='none'` regardless of the
    actual value of `partial_policy`, unless `partial_policy` is
    ``'replace'``, which completely ignores the limit.
    """
    if partial_policy is not None:
        partial_type = check_value(partial_policy, ensure_extension.policies,
                                   label='match policy')
    else:
        partial_type = None

    if name.endswith(ext):
        # This should handle the empty `ext` case as well
        return name

    # At this point, `ext` is guarantee to be non-empty
    index = name.rfind(ext[0])

    if index == -1 or partial_type in (None, '', 'none'):
        return name + ext
    # At this point, `name` is guaranteed to contain `ext[0]` at index

    if partial_type in ('create', 'c'):
        return name

    def right_size(longer=True):
        """
        True if `partial_limit` is `None`, or if `ext` at least as long
        as the suffix of `name`, but by no more than `partial_limit`
        characters. This means that the name can be appended/inserted to
        to get the desired extension.

        If `longer` is `False`, the sense of the comparison is reversed:
        `ext` must be shorter than the suffix so the name can be
        stripped/removed from to get the extension.
        """
        delta = len(ext) - (last - index)
        if not longer:
            delta = -delta
        return partial_limit is None or 0 <= delta <= partial_limit

    last = len(name) - 1
    replace = False
    if partial_type in ('append', '+'):
        if right_size(True) and ext.startswith(name[index:]):
            replace = True
    elif partial_type in ('insert', '^'):
        if right_size(True) and _is_ordered_subset(name[index:], ext):
            replace = True
    elif partial_type in ('strip', '-'):
        if right_size(False) and name[index:].startswith(ext):
            replace = True
    elif partial_type in ('remove', 'x'):
        if right_size(False) and _is_ordered_subset(ext, name[index:]):
            replace = True
    else: # partial_type == 'replace'
        replace = True

    if replace:
        return name[:index] + ext
    return name + ext


ensure_extension.policies = (
    'none', '', 'append', '+', 'insert', '^', 'strip', '-', 'remove', 'x',
    'replace', 'r', 'create', 'c',
)


def split_extension(filename, max=None):
    """
    Upgraded version of :py:func:`os.path.splitext` that splits apart
    all the available extensions.

    Parameters
    ----------
    filename : str or ~pathlib.Path
        The file name to split.
    max : int or None, optional
        The maximum number of extensions to split off. ``max=1`` is
        equivalent to calling `os.path.splitext`. ``max=None``, zero
        or negative values means split all extensions off. Default
        is `None`.

    Returns
    -------
    parts : list
        A list containing the base name and up to `max` extensions.
        Any unsplit extensions will still be attached to the base name.
    """
    if max is None or max <= 0:
        max = -1

    parts = [filename]
    while max != 0:
        s = splitext(parts[0])
        if not s[1]:
            break
        parts[:1] = s
        max -= 1
    return parts


def insert_suffix(filename, suffix, n=0, allow_duplicate=False):
    """
    Insert a suffix into the file name before the extension.

    Append the suffix if there is no extension. By default, if the
    suffix is already present, it is not duplicated.

    Parameters
    ----------
    filename : str or ~pathlib.Path
        The name to modify.
    suffix : str
        The suffix to insert.
    n : int, optional
        The extension after which to insert the suffix. Indexing
        similar to list indexing, with ``n=0`` referring to the base
        name and ``n=-1`` the last extension. In ``"a.b.c.d"``,
        ``".d"`` is at ``n=3`` or ``n=-1``, ``"a"`` is at ``n=0`` or
        ``n=-4``. The default is to prepend to the base name: ``n=0``.
    allow_duplicate : bool
        If :py:obj:`True`, no check will be made to see if the suffix is
        already present. If :py:obj:`False` (the default), the suffix
        will only be inserted if not already present.

    Return
    ------
    inserted : str
        The modified name.
    """
    parts = split_extension(filename)
    if allow_duplicate or not parts[n].endswith(suffix):
        parts[n] += suffix
        return ''.join(parts)
    return str(filename)


class PreOpenedFile:
    """
    A proxy class for file objects that does not open or close the file
    when :py:meth:`__enter__`, :py:meth:`__exit__` and :py:meth:`close`
    are invoked.

    This version of the usual context manager is useful when processing
    opened files along with strings.

    Note that this is not a general-purpose proxy that can be used for
    most objects because it does not define any special methods besides
    :py:meth:`__repr__`, :py:meth:`__enter__`, :py:meth:`__exit__`,
    :py:meth:`__iter__` and :py:meth:`close` as class attributes.
    """
    def __init__(self, file):
        self.__file = file

    def __getattr__(self, name):
        """
        Pass through access to all missing attributes to the underlying
        file object.
        """
        return getattr(self.__file, name)

    def __repr__(self):
        """
        Return a string representation of the underlying file.
        """
        return '{}({})'.format(__class__.__name__, repr(self.__file))

    def __enter__(self, *args, **kwargs):
        """ Return this proxy object. """
        return self

    def __exit__(self, *args, **kwargs):
        """ Do nothing. """
        pass

    def __iter__(self, *args, **kwargs):
        """ Return an iterator over the underlying file. """
        return iter(self.__file)

    def close(self):
        """ Do nothing. """
        pass


def open_file(file, *args, **kwargs):
    """
    Return an open file-like object for the input.

    If the input is already a file-like object (not a string, path, or
    file descriptor), a proxy for it is returned. The original object
    remains unmodified. The proxy can be used in a context manager, but
    it will not close the file when exiting.

    Strings, paths and file descriptors are opened using the additional
    arguments provided. They return a true file object that will close
    itself when used as a context manager.

    Returns
    -------
    file : file-like
        A file-like opened from the `file` input.
    """
    # Checking for file vs descriptor
    try:
        return open(file, *args, **kwargs)
    except TypeError:
        return PreOpenedFile(file)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 01 Nov 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.logs` module.
"""


from contextlib import contextmanager
from io import StringIO
import logging
import sys
from warnings import catch_warnings

from pytest import fixture, raises

from ..logs import (
    KEEP, KEEP_WARN, OVERWRITE, OVERWRITE_WARN, RAISE,
    add_logging_level, reset_handlers, LogMaxFilter, MetaLoggableType
)


class LoggingTestBase(metaclass=MetaLoggableType):
    levelNames = {
        lvl: logging.getLevelName(lvl) for lvl in (
            logging.DEBUG, logging.INFO, logging.WARNING,
            logging.ERROR, logging.CRITICAL
        )
    }

    @fixture
    def reset_handler(self):
        return True

    @fixture
    def initial_level(self):
        return logging.DEBUG

    @fixture
    def formatter(self):
        return logging.Formatter('%(levelname)s %(message)s')

    @fixture(autouse=True)
    def log_setup(self, initial_level, reset_handler, formatter):
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        if reset_handler:
            reset_handlers(self.handler, logger=self.logger,
                           level=initial_level, format=formatter,
                           filter_type=False)
        else:
            self.handler.setFormatter(formatter)
            self.handler.setLevel('NOTSET')
            self.logger.addHandler(self.handler)
        self.logger.setLevel(initial_level)


class TestAddLoggingLevel(LoggingTestBase):
    @contextmanager
    def remove_level(self, name, method=None):
        yield
        if method is None:
            method = name.lower()
        logging._acquireLock()
        try:
            level = getattr(logging, name)
            delattr(logging, name)
            delattr(logging, method)
            delattr(logging.getLoggerClass(), method)
            delattr(logging.LoggerAdapter, method)
            if hasattr(logging, '_levelToName') and \
                                level in logging._levelToName:
                del logging._levelToName[level]
            if hasattr(logging, '_nameToLevel') and \
                                name in logging._nameToLevel:
                del logging._nameToLevel[name]
        finally:
            logging._releaseLock()

    def test_basic(self):
        with self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5)
            assert getattr(logging, 'XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName('XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName(logging.DEBUG + 5) == 'XDEBUG'
            assert hasattr(logging, 'xdebug')
            assert hasattr(logging.getLoggerClass(), 'xdebug')
            assert hasattr(logging.LoggerAdapter, 'xdebug')

    def test_keep(self):
        with catch_warnings(record=True) as w, self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=KEEP)
            assert not w

            add_logging_level('XDEBUG', logging.DEBUG + 1, if_exists=KEEP)
            assert not w

            assert getattr(logging, 'XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName('XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName(logging.DEBUG + 5) == 'XDEBUG'
            assert hasattr(logging, 'xdebug')
            assert hasattr(logging.getLoggerClass(), 'xdebug')
            assert hasattr(logging.LoggerAdapter, 'xdebug')

    def test_keep_warn(self):
        with catch_warnings(record=True) as w, self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=KEEP_WARN)
            assert not w

            add_logging_level('XDEBUG', logging.DEBUG + 1, if_exists=KEEP_WARN)
            assert len(w) == 1
            assert 'XDEBUG' in str(w[0].message)
            assert '2/5' in str(w[0].message)  # logging.XDEBUG and mapping

            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=KEEP_WARN)
            assert len(w) == 1

            assert getattr(logging, 'XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName('XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName(logging.DEBUG + 5) == 'XDEBUG'
            assert hasattr(logging, 'xdebug')
            assert hasattr(logging.getLoggerClass(), 'xdebug')
            assert hasattr(logging.LoggerAdapter, 'xdebug')

    def test_overwrite(self):
        with catch_warnings(record=True) as w, self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=OVERWRITE)
            assert not w

            add_logging_level('XDEBUG', logging.DEBUG + 1, if_exists=OVERWRITE)
            assert not w

            assert getattr(logging, 'XDEBUG') == logging.DEBUG + 1
            assert logging.getLevelName('XDEBUG') == logging.DEBUG + 1
            assert logging.getLevelName(logging.DEBUG + 5) == 'XDEBUG'
            assert hasattr(logging, 'xdebug')
            assert hasattr(logging.getLoggerClass(), 'xdebug')
            assert hasattr(logging.LoggerAdapter, 'xdebug')

    def test_overwrite_warn(self):
        with catch_warnings(record=True) as w, self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5,
                              if_exists=OVERWRITE_WARN)
            assert not w

            add_logging_level('XDEBUG', logging.DEBUG + 1,
                              if_exists=OVERWRITE_WARN)
            assert len(w) == 1
            assert 'XDEBUG' in str(w[0].message)
            assert '2/5' in str(w[0].message)  # logging.XDEBUG and mapping

            add_logging_level('XDEBUG', logging.DEBUG + 1,
                              if_exists=OVERWRITE_WARN)
            assert len(w) == 1

            assert getattr(logging, 'XDEBUG') == logging.DEBUG + 1
            assert logging.getLevelName('XDEBUG') == logging.DEBUG + 1
            assert logging.getLevelName(logging.DEBUG + 5) == 'XDEBUG'
            assert hasattr(logging, 'xdebug')
            assert hasattr(logging.getLoggerClass(), 'xdebug')
            assert hasattr(logging.LoggerAdapter, 'xdebug')

    def test_raise(self):
        with catch_warnings(record=True) as w, self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=RAISE)
            assert not w

            raises(AttributeError, add_logging_level, 'XDEBUG',
                   logging.DEBUG + 1, if_exists=RAISE)
            assert not w

            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=RAISE)
            assert not w

            assert getattr(logging, 'XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName('XDEBUG') == logging.DEBUG + 5
            assert logging.getLevelName(logging.DEBUG + 5) == 'XDEBUG'
            assert hasattr(logging, 'xdebug')
            assert hasattr(logging.getLoggerClass(), 'xdebug')
            assert hasattr(logging.LoggerAdapter, 'xdebug')

    def test_logger(self):
        with self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=RAISE)
            self.logger.xdebug('message1')
            self.logger.log(logging.XDEBUG, 'message2')
        self.handler.flush()
        assert self.stream.getvalue() == 'XDEBUG message1\nXDEBUG message2\n'

    def test_logger_adapter(self):
        with self.remove_level('XDEBUG'):
            add_logging_level('XDEBUG', logging.DEBUG + 5, if_exists=RAISE)
            adapter = logging.LoggerAdapter(self.logger, {})
            adapter.xdebug('message1')
            adapter.log(logging.XDEBUG, 'message2')
        self.handler.flush()
        assert self.stream.getvalue() == 'XDEBUG message1\nXDEBUG message2\n'

    def test_bad_name(self):
        raises(ValueError, add_logging_level, 'lowercase', 1)


class TestResetHandlers(LoggingTestBase):
    @fixture
    def reset_handler(self):
        return False

    def test_handler_none(self):
        raises(AttributeError, reset_handlers, None, logger=self.logger)

    def test_format(self):
        stream = StringIO()
        if sys.version_info >= (3, 8):
            format = logging.Formatter('XXX', validate=False)
        else:
            format = logging.Formatter('XXX')
        reset_handlers(logging.StreamHandler(stream), logger=self.logger,
                       format=format, filter_type=False)
        self.logger.debug('test')
        assert len(self.logger.handlers) == 1
        self.logger.handlers[0].flush()
        assert stream.getvalue() == 'XXX\n'

    def test_logger(self):
        logger = logging.getLogger(self.logger.name + '.a')

        for i in range(3):
            self.logger.addHandler(logging.NullHandler())

        reset_handlers(logging.NullHandler(), logger=logger, filter_type=False)

        assert len(self.logger.handlers) >= 3
        assert len(logger.handlers) == 1

    def test_level(self):
        reset_handlers(logging.NullHandler(), logger=self.logger,
                       level='WARNING', filter_type=False)
        assert len(self.logger.handlers) == 1
        assert self.logger.handlers[0].level == logging.WARNING

    def test_remove_hook(self):
        for i in range(3):
            self.logger.addHandler(logging.NullHandler())

        def remove_hook(handler):
            nonlocal remove_count
            remove_count += 1

        intial_count = len(self.logger.handlers)
        remove_count = 0

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       remove_hook=remove_hook, filter_type=False)

        assert intial_count >= 3
        assert remove_count == intial_count

    def test_bad_filter_type(self):
        raises(TypeError, reset_handlers, logging.NullHandler(),
               filter_type='a')
        raises(TypeError, reset_handlers, logging.NullHandler(),
               filter_type=logging.StreamHandler)

    def test_type_filter_type(self):
        self.logger.handlers[:] = []
        for i in range(3):
            self.logger.addHandler(logging.StreamHandler(StringIO()))
            self.logger.addHandler(logging.NullHandler())
        assert len(self.logger.handlers) == 6

        reset_handlers(logging.NullHandler(), logger=self.logger)
        assert len(self.logger.handlers) == 4
        assert sum(isinstance(h, logging.NullHandler)
                   for h in self.logger.handlers) == 1
        assert sum(isinstance(h, logging.StreamHandler)
                   for h in self.logger.handlers) == 3

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       filter_type=logging.Handler)
        assert len(self.logger.handlers) == 1
        assert isinstance(self.logger.handlers[0], logging.NullHandler)

    def test_true_filter_type(self):
        self.logger.handlers[:] = []
        for i in range(3):
            self.logger.addHandler(logging.StreamHandler(StringIO()))
            self.logger.addHandler(logging.NullHandler())
        assert len(self.logger.handlers) == 6

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       filter_type=True)
        assert len(self.logger.handlers) == 4
        assert sum(isinstance(h, logging.NullHandler)
                   for h in self.logger.handlers) == 1
        assert sum(isinstance(h, logging.StreamHandler)
                   for h in self.logger.handlers) == 3

    def test_false_filter_type(self):
        self.logger.handlers[:] = []
        for i in range(3):
            self.logger.addHandler(logging.StreamHandler(StringIO()))
            self.logger.addHandler(logging.NullHandler())
        assert len(self.logger.handlers) == 6

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       filter_type=False)
        assert len(self.logger.handlers) == 1
        assert isinstance(self.logger.handlers[0], logging.NullHandler)

    def test_filter_hook_no_type(self):
        self.logger.handlers[:] = []
        for i in range(3):
            self.logger.addHandler(logging.StreamHandler(StringIO()))
            self.logger.addHandler(logging.NullHandler())
        assert len(self.logger.handlers) == 6

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       filter_hook=lambda h: True)
        assert len(self.logger.handlers) == 1
        assert isinstance(self.logger.handlers[0], logging.NullHandler)

    def test_filter_hook(self):
        self.logger.handlers[:] = []
        for i in range(6):
            htype = type('Handler{}'.format(i), (logging.Handler,), {})
            handler = htype(level=10 * i)
            self.logger.addHandler(handler)
        assert len(self.logger.handlers) == 6

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       filter_hook=lambda h: h.level < 30)
        assert len(self.logger.handlers) == 4
        assert sum(isinstance(h, logging.NullHandler)
                   for h in self.logger.handlers) == 1

    def test_filter_hook_and_type(self):
        self.logger.handlers[:] = []
        for i in range(6):
            htype = type('Handler{}'.format(i), (logging.Handler,), {})
            handler = htype(level=10 * i)
            self.logger.addHandler(handler)
        assert len(self.logger.handlers) == 6

        reset_handlers(logging.NullHandler(), logger=self.logger,
                       filter_hook=lambda h: h.level < 30,
                       filter_type=logging.NullHandler)
        assert len(self.logger.handlers) == 7
        assert sum(isinstance(h, logging.NullHandler)
                   for h in self.logger.handlers) == 1

    def test_bad_filter_hook(self):
        raises(TypeError, reset_handlers, logger=self.logger, filter_hook='a')


class TestLogMaxFilter(LoggingTestBase):
    @fixture
    def reset_handler(self):
        return False

    @contextmanager
    def temp_log_max_filter(self, obj, *args, **kwargs):
        filter = LogMaxFilter(*args, **kwargs)
        obj.addFilter(filter)
        yield
        obj.removeFilter(filter)
        self.handler.flush()

    def test_inclusive_logger(self):
        with self.temp_log_max_filter(self.logger, 'WARNING', inclusive=True):
            for value, name in self.levelNames.items():
                self.logger.log(value, name)

        text = self.stream.getvalue()
        for value, name in self.levelNames.items():
            if value >= logging.WARNING:
                assert name not in text
            else:
                assert ' '.join((name, name)) in text

    def test_exclusive_logger(self):
        with self.temp_log_max_filter(self.logger, logging.WARNING,
                                      inclusive=False):
            for value, name in self.levelNames.items():
                self.logger.log(value, name)

        text = self.stream.getvalue()
        for value, name in self.levelNames.items():
            if value > logging.WARNING:
                assert name not in text
            else:
                assert ' '.join((name, name)) in text

    def test_inclusive_handler(self):
        with self.temp_log_max_filter(self.handler, logging.WARNING,
                                      inclusive=True):
            for value, name in self.levelNames.items():
                self.logger.log(value, name)

        text = self.stream.getvalue()
        for value, name in self.levelNames.items():
            if value >= logging.WARNING:
                assert name not in text
            else:
                assert ' '.join((name, name)) in text

    def test_exclusive_handler(self):
        with self.temp_log_max_filter(self.handler, 'WARNING',
                                      inclusive=False):
            for value, name in self.levelNames.items():
                self.logger.log(value, name)

        text = self.stream.getvalue()
        for value, name in self.levelNames.items():
            if value > logging.WARNING:
                assert name not in text
            else:
                assert ' '.join((name, name)) in text

        self.handler.removeFilter(filter)


    def test_custom_level(self):
        with self.temp_log_max_filter(self.logger, logging.WARNING + 5,
                                      inclusive=False):
            for value, name in self.levelNames.items():
                self.logger.log(value, name)

        text = self.stream.getvalue()
        for value, name in self.levelNames.items():
            if value > logging.WARNING:
                assert name not in text
            else:
                assert ' '.join((name, name)) in text

    def test_bad_level(self):
        raises(TypeError, LogMaxFilter, '_NOTALEVEL32', inclusive=True)
        raises(TypeError, LogMaxFilter, '_NOTALEVEL32', inclusive=False)


class TestMetaLoggableType:
    def test_regular(self):
        class Test(metaclass=MetaLoggableType):
            namespace = 'xxx'

        expected_name = Test.__module__ + '.' + Test.__qualname__

        assert not hasattr(Test, '__namespace__')
        assert isinstance(Test.logger, logging.Logger)
        assert Test.logger.name == expected_name
        assert logging.getLogger(expected_name) is Test.logger

    def test_namespace(self):
        class Test(metaclass=MetaLoggableType):
            __namespace__ = 'xxx'

        expected_name = Test.__namespace__ + '.' + Test.__module__ + '.' + Test.__qualname__

        assert hasattr(Test, '__namespace__')
        assert isinstance(Test.logger, logging.Logger)
        assert Test.logger.name == expected_name
        assert logging.getLogger(expected_name) is Test.logger

    def test_builtin(self):
        """ Verify that builtins play nicely with the metaclass. """
        class Test(list, metaclass=MetaLoggableType):
            __namespace__ = 'xyz'

        expected_name = Test.__namespace__ + '.' + Test.__module__ + '.' + Test.__qualname__

        assert hasattr(Test, '__namespace__')
        assert isinstance(Test.logger, logging.Logger)
        assert Test.logger.name == expected_name
        assert logging.getLogger(expected_name) is Test.logger



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
:py:mod:`haggis.version` holds the version of the package.

The actual version number is normally exported through the top-level
:py:mod:`haggis` package, but defined here. This module exists to
facilitate deployment and testing, so users do not need to install the
entire package to get the version number.

Bump the version in a feature branch before merging. Unless a major
rewrite is planned, the new version should increment the minor number.

The version in a feature branch should be a release candidate, and have
the corresponding `rc1` suffix. Bump to full version immediately before
merging with master. Most commits in master should have a tag.
"""

#: The current version
__version__ = '0.13.0'



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 09 Jan 2021: Initial Coding

"""
Pure-python implementation of some useful data structures.
"""


__all__ = ['Trie']


from collections import deque
from itertools import chain
from os import sep
from os.path import isabs


class _TrieNode:
    """
    Contains information about a Node.

    This class has four slots:

    .. py:attribute:: key

       The key identifying this node.

    .. py:attribute:: parent

       A reference to the parent node for ease of traversal.

    .. py:attribute:: isleaf

       A boolean indicating whether this is a leaf or not. Since a leaf
       marks the end of a complete string, leaves can occur in the
       middle of a :py:class:`Trie` when a prefix is a valid string in
       its own right.

    .. py:attribute:: suffixes

       A mapping of suffix nodes prefixed by the current node, keyed by
       their :py:attr:`key` attribute. May be `None` to indicate an
       empty node.
    """
    __slots__ = ('key', 'parent', 'isleaf', 'suffixes')

    def __init__(self, key, parent, isleaf=False):
        """
        Initialize a new empty node.

        Inputs are copied directly, and :py:attr:`suffixes` is set to
        `None`.

        Parameters
        ----------
        key :
            The :py:attr:`key` to assign.
        parent : _TrieNode
            The :py:attr:`parent` node.
        isleaf : bool, optional
            The default :py:attr:`isleaf` is `False`.
        """
        self.key = key
        self.parent = parent
        self.isleaf = isleaf
        self.suffixes = None

    def __bool__(self):
        """
        The truth value of a node, used as a convenience when deleting
        nodes.

        Only nodes that are not leaves and empty are `False`. A root
        node, with a `None` parent, may not be `False`.
        """
        return self.isleaf or not self.isempty or self.parent is None

    def __str__(self):
        """
        Return a description of this node that will be used to display
        it as part of the entire structure.

        Returns
        -------
        str
            The key, followed by an asterisk if it is a leaf.
        """
        s = repr(self.key)
        if self.isleaf:
            s += '*'
        return s

    def __repr__(self):
        """
        Create a programatic description of this instance.

        Returns
        -------
        str
            A string that can be used to approximate a constructor call,
            with no information about suffixes.
        """
        return '{}(key={!r}, parent={}@{}, isleaf={})'.format(
                type(self).__name__, self.key, type(self.parent).__name__,
                id(self.parent), self.isleaf
        )

    def __len__(self):
        """
        The number of suffixes in this node.

        Returns
        -------
        int
            The number of child nodes.
        """
        return 0 if self.isempty else len(self.suffixes)

    def __contains__(self, key):
        """
        Check if the specified key is a valid suffix for this node.

        Parameters
        ----------
        key : str
            The key to check for.

        Returns
        -------
        bool
            `True` if this node has a suffix with the specified key,
            `False` otherwise.
        """
        return False if self.isempty else key in self.suffixes

    def __getitem__(self, key):
        """
        Retrieve a suffix node, creating it if necessary.

        This method accepts a single key: it is non-recursive.

        Parameters
        ----------
        key : str
            The key of the suffix to retrieve.

        Returns
        -------
        TrieNode
            The suffix node with the specified key. If necessary, a
            non-leaf node will be created for the given key, with
            this node set as the parent.
        """
        if self.isempty:
            self.suffixes = {}
        if key not in self.suffixes:
            self.suffixes[key] = type(self)(key, self)
        return self.suffixes[key]

    def __delitem__(self, key):
        """
        Remove the specified key from this node.

        Missing keys are silently ignored.

        Parameters
        ----------
        key : str
            The key corresponding to the suffix to remove. May be
            missing.
        """
        if key in self:
            if len(self.suffixes) == 1:
                self.suffixes = None
            else:
                del self.suffixes[key]

    @property
    def keylist(self):
        """
        Retrieves a list of all the suffix keys in this node.

        Read-only property.
        """
        if self.isempty:
            return []
        return list(self.suffixes.keys())

    @property
    def isempty(self):
        """
        `True` if this node has suffixes.

        Read-only property.
        """
        return self.suffixes is None

    @property
    def hierarchy(self):
        """
        Retrieve all the nodes between the root and this one.

        Read-only property.
        """
        p = []
        node = self
        while node is not None:
            p.append(node.key)
            node = node.parent
        return p[::-1]


class Trie:
    """
    Simple general purpose Trie implementation with methods for adding,
    removing, checking containment, and iterating.

    Root can be a leaf if the empty string is a prefix. Keys must be
    hashable.
    """
    __slots__ = ('root', 'len', 'sorter', 'joiner')

    def __init__(self, empty=None, sorter=None, joiner=None):
        """
        Make an empty trie.

        Do not delete existing data (e.g. if invoked multiple times).

        Parameters
        ----------
        empty : optional
            The key of the root node, associated with an empty trie.
            The default is `None`.
        sorter : callable, optional
            A callable that determines the sort order of suffixes
            during default iteration. If provided, must accept a list
            of keys and return an iterable. May filter the input by
            returning a subset. May be `None` (the default) to indicate
            no sorting on iteration. See :py:meth:`iter` for more
            information.
        joiner : callable, optional
            A callable that determines the concatenation of nodes
            during default iteration. If provided, must accept an
            iterable of keys and return the concatenated object. May
            be `None` (the default) to indicate no concatenation. See
            :py:meth:`iter` for more information.
        """
        if not hasattr(self, 'root'):
            self.root = _TrieNode(empty, None)
            self.len = 0
            self.sorter = (lambda x: x) if sorter is None else sorter
            self.joiner = tuple if joiner is None else joiner

    def add(self, item):
        """
        Add an item, represented as an iterable of parts.

        The last element of the iterable is marked as a leaf.

        Parameters
        ----------
        item :
            An iterable of keys. The last element will be marked as a
            leaf. An empty iterable refers to the root node.

        Returns
        -------
        bool
            `True` if a new leaf is added (even if it is prefix to an
            existing suffix), `False` if already a leaf.
        """
        node = self.root
        for elem in item:
            node = node[elem]
        if node.isleaf:
            return False
        self.len += 1
        node.isleaf = True
        return True

    def remove(self, item):
        """
        Remove an item, represented as an iterable of parts.

        If the leaf has a suffix, it is simply unmarked. If not, it, and
        its parents will be removed until a node with a different suffix
        or that is a leaf is encountered.

        Parameters
        ----------
        item :
            An iterable of keys. The last element will no longer be a
            leaf and may be deleted. An empty iterable refers to the
            root node.

        Returns
        -------
        bool
            `True` if the item was found and a node was removed, `False`
            if it did not represent a valid leaf.
        """
        node = self.root
        for elem in item:
            if elem not in node:
                return False
            node = node[elem]
        if node.isleaf:
            node.isleaf = False
            self.len -= 1
            while not node:
                del node.parent[node.key]
                node = node.parent
            return True
        return False

    def __len__(self):
        """
        Length is the number of leaf nodes, consistent with iteration.

        Returns
        -------
        int
            The number of leaves in this trie.
        """
        return self.len

    def __contains__(self, item):
        """
        Check if the specified item, represented as an iterable of
        parts, is a leaf of this trie.

        Parameters
        ----------
        item :
            An iterable of keys. The last element is checked for
            leafness. An empty iterable refers to the root node.

        Returns
        -------
        bool
            `True` if a the last element of `item` represents a leaf
            in the trie, `False` otherwise, even if it is a valid node.
        """
        node = self.root
        for elem in item:
            if not elem in node:
                return False
            node = node[elem]
        return node.isleaf

    def __iter__(self):
        """
        Default iterator over the leaf sequences.

        Iteration happens in deapth-first order. A customizable iterator
        is available through the :py:meth:`iter` method.
        """
        return self.iter()

    def iter(self, sorter=None, joiner=None, dfs=True):
        """
        Custom iterator over leaves of the trie.

        Parameters
        ----------
        sorter : callable or None
            If provided, used to sort the suffixes for each node.
            Otherwise, leaves will appear in insertion order (or none at
            all pre-Python 3.6). `sorter` must be a callable that
            accepts a list of keys and returns an iterable of keys. It
            can be used to filter elements as well as sort them.
        joiner : callable
            If provided, used to concatenate all the elements of each
            leaf. Otherwise, yield tuples with the elements. `joiner`
            must be a callable that accepts an iterable of keys and
            returns the concatenated object. The first element will always
            be the empty root key.
        dfs : bool
            Whether to perform breadth-first or depth-first-search. If
            `dfs` is False, a breadth-first order will be used rather
            than depth-first.
        """
        if sorter is None:
            sorter = self.sorter
        if joiner is None:
            joiner = self.joiner

        if dfs:
            def dfs_node(node):
                # Don't modify in place
                if node.isleaf:
                    yield node
                for key in sorter(node.keylist):
                    yield from dfs_node(node[key])
            yield from (joiner(node.hierarchy) for node in dfs_node(self.root))
        else:
            stack = deque([self.root])
            while stack:
                node = stack.popleft()
                if node.isleaf:
                    yield joiner(node.hierarchy)
                stack.extend(node[suffix] for suffix in sorter(node.keylist))

    def __repr__(self):
        """
        String representation of this trie.

        Returns
        -------
        str
            A multi-line description, with nodes indented to show
            hierarchy.
        """
        s = []
        def _repr(node, i):
            s.append(' ' * i + repr(node))
            for key in self.sorter(node.keylist):
                _repr(node[key], i + 2)
        _repr(self.root, 2)
        return '{}@{}{}{}'.format(type(self).__name__, id(self),
                                  '\n', '\n'.join(s))

    @classmethod
    def strings(cls):
        """
        Create a trie for strings.

        The root is an empty string, `sorter` is :py:func:`sorted`, and
        `jointer` is ``''.join``.

        Returns
        -------
        Trie
            A trie for strings.
        """
        return cls(empty='', sorter=sorted, joiner=''.join)

    @classmethod
    def string_paths(cls, sorter=sorted, joiner=None):
        """
        Create a trie for paths.

        Able to handle relative and absolute paths fairly well in the
        same trie,

        Parameters
        ----------
        sorter : callable, optional
            A replacement sorter. The default is :py:func:`sorted`, which
            implies case sensitivity.
        joiner : callable, optional
            A replacement joiner. The default joiner handles
            concatenation with the correct path separator, and proper
            identification of absolute and relative paths.

        Returns
        -------
        Trie
            A trie for paths.
        """
        if joiner is None:
            def joiner(parts):
                parts = iter(parts)
                root = next(parts)
                first = next(parts, None)
                if first is None:
                    return sep
                elif first == '.':
                    prefix = '.'
                elif isabs(first):
                    # Handle C:\\ on Windows
                    prefix = [first.rstrip('\\')]
                else:
                    prefix = [root, first]
                return sep.join(chain(prefix, parts))
        return cls(empty='', sorter=sorter, joiner=joiner)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 11 Jan 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.files.csv` module.
"""


import numpy
from pytest import raises

from ..csv import reformat


class TestReformat:
    def setup_method(self):
        self.data = [[1, 'a'], [2, 'b'], [3, 'c']]

    def test_normal(self):
        assert reformat(self.data, 'normal') is self.data

    def test_transpose(self):
        assert reformat(self.data, 'transpose') == [[1, 2, 3], ['a', 'b', 'c']]

    def test_numpy(self):
        result = reformat(self.data, 'numpy')
        assert isinstance(result, numpy.ndarray)
        assert result.shape == (len(self.data), len(self.data[0]))
        numpy.testing.assert_array_equal(result, numpy.array(self.data))

    def test_invalid(self):
        with raises(ValueError):
            reformat(self.data, 'invalid')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for handling MatPlotLib figures, only available when the
``[plot]`` :ref:`extra <installation-extras>` is installed.

If `matplotlib`_ is not found at import time, this module will have only
a :py:data:`plot_enabled` attribute, which will be :py:obj:`False`. If
`matplotlib`_ is found, on the other hand, :py:data:`plot_enabled` will
be :py:obj:`True`, and all the other documented functions and attributes
of the module will be present.

.. py:data:: plot_enabled

   A boolean value indicating whether the ``[plot]``
   :ref:`extra <installation-extras>` has been installed. If
   :py:obj:`False`, the API will be severely limited.


.. include:: /link-defs.rst 
"""

__all__ = ['plot_enabled']


from contextlib import contextmanager
from io import BytesIO


try:
    from matplotlib import pyplot as plt
    from matplotlib import ticker as mtick
except ImportError:
    from . import _display_missing_extra
    _display_missing_extra('plot', 'matplotlib')
    plot_enabled = False
else:
    import numpy

    __all__.extend([
        'figure_context', 'save_figure', 'set_figure_size', 'set_labels',
        'show_extents', 'semilogx_base', 'semilogy_base', 'loglog_base',
    ])
    plot_enabled = True


if plot_enabled:
    @contextmanager
    def figure_context(*args, **kwargs):
        """
        A context manager that automatically closes the figure that it
        opens.

        Inspired by https://github.com/matplotlib/matplotlib/issues/5218/#issue-110729876
        """
        fig = plt.figure(*args, **kwargs)
        yield fig
        plt.close(fig)


    def save_figure(fig, file=None, size=None, **kwargs):#dpi=300, format='png'
        """
        Save the figure as an image using
        :py:meth:`matplotlib.figure.Figure.savefig`.

        The main value of this method is that it automatically saves to
        memory via a :py:class:`~io.BytesIO` object if a file is not
        specified.

        Parameters
        ----------
        fig:
            The figure to save
        file: str, file-like, or None
            If not :py:obj:`None`, there is no return value. If
            :py:obj:`None`, :py:class:`~io.BytesIO` containing the image
            will be returned. The output will be rewound to the start in
            that case. The default is :py:obj:`None`.
        size: sequence[int]
            The size of the figure in inches, as a 2-element sequence
            ``(w, h)``. If either of the elements is :py:obj:`None`, the
            aspect ratio of the figure will be preserved. Even if only
            the width is specified, `size` must be a sequence.


        All other arguments are passed through directly to
        :py:meth:`~matplotlib.figure.Figure.savefig`. Some common
        options include:

        dpi :
            The resolution of the output image in dots-per-inch.
        format :
            The output format. ``'png'``, ``'svg'``, ``'pdf'`` have good
            support. If not supplied, the default is explicitly set to
            ``'png'``.
        frameon :
            Whether of nor the figure background should be rendered.
            Defaults to :py:obj:`True` if not supplied.
        transparent :
            Whether or not the axes background should be rendered as
            transparent. Defaults to the inverse value of `frameon` if
            not supplied.
        bbox_inches :
            The portion of the figure to save. If ``'tight'``, try to
            use the entire figure. If unset, defaults to ``'tight'``.
        pad_inches :
            The amount of padding to add around the figure when
            `bbox_inches` is ``'tight'``.
        """
        output = BytesIO() if file is None else file
        if size is not None:
            set_figure_size(fig, *size)
        kwargs.setdefault('format', 'png')
        kwargs.setdefault('bbox_inches', 'tight')
        kwargs.setdefault('frameon', fig.frameon)
        kwargs.setdefault('transparent', not kwargs.get('frameon'))
        fig.savefig(output, **kwargs)
        if file is None:
            output.seek(0)
        return output


    def set_figure_size(fig, w=None, h=None):
        """
        Set the size of a figure in inches, optionally preserving the
        aspect ratio.

        If either or the size arguments is :py:obj:`None`, it will be
        scaled to preserve the current aspect ratio. If both are
        :py:obj:`None`, the size is not set at all.
        """
        curr_w, curr_h = fig.get_size_inches()
        if w is None:
            if h is None:
                return
            w = h * curr_w / curr_h
        elif h is None:
            h = w * curr_h / curr_w
        fig.set_size_inches(w, h)


    def set_labels(artists, labels):
        """
        Assign a separate label to each artist in the iterable.

        Useful in labelling each column separately when plotting a
        multi-column array. For example::

            from matplotlib import pyplot as plt
            import numpy as np

            x = np.arange(5)
            y = np.random.ranint(10, size=(5, 3))

            fig, ax = plt.subplots()
            set_labels(ax.plot(x, y), 'ABC')

        Based on https://stackoverflow.com/a/64780035/2988730.

        Parameters
        ----------
        artists :
            Iterable of artists. Any extra entries are silently ignored
            (not labeled).
        labels :
            Iterable of strings. Any extra labels are silently dropped.
        """
        for artist, label in zip(artists, labels):
            artist.set_label(label)


    def show_extents(img, x=None, y=None, ax=None, **kwargs):
        """
        Display an image with the correct x- and y- coordinates,
        adjusted to pixel centers.

        This function is a wrapper around
        :py:meth:`~matplotlib.axes.Axes.imshow`. Normally, ``imshow``
        will scale the axes limits to the outer edges of the image
        when given an ``extent`` argument. However, it is generally
        more accurate to set the centers of the pixels.

        Parameters
        ----------
        img :
            The image to display.
        x : array-like, optional
            The x-coordinates of the pixels. Only the first and last
            coordinate are ever used, so it is safe to pass in any
            sequence of two numbers. ``x[0]`` is the intended
            x-coordinate of the center of the leftmost column of the
            image, while ``x[-1]`` is the x-coordinate of the center of
            the rightmost column.

            Defaults to ``[0, img.shape[1] - 1]``.
        y : array-like, optional
            The y-coordinates of the pixels. Only the first and last
            coordinate are ever used, so it is safe to pass in any
            sequence of two numbers. ``y[0]`` is the intended
            y-coordinate of the center of the topmost row of the image,
            while ``y[-1]`` is the y-coordinate of the center of
            the bottommost column.

            Defaults to ``[0, img.shape[0] - 1]``.
        ax : matplotlib.axes.Axes, optional
            The axes to plot on. If not supplied, a new figure and axes
            are created.
        **kwargs :
            All remaining arguments are passed through to
            :py:meth:`~matplotlib.axes.Axes.imshow`. If an explicit
            ``extent`` is passed in, ``x`` and ``y`` will be ignored.

        Return
        ------
        image : matplotlib.image.AxesImage
            The image object created by
            :py:meth:`~matplotlib.axes.Axes.imshow`.
        """
        if ax is None:
            _, ax = plt.subplots()

        if 'extent' in kwargs:
            extent = kwargs.pop('extent')
        else:
            if x is None:
                x = [0, img.shape[1] - 1]
            else:
                x = numpy.asanyarray(x).ravel()
            if y is None:
                y = [0, img.shape[0] - 1]
            else:
                y = numpy.asanyarray(y).ravel()

            dx = 0.5 * (x[-1] - x[0]) / (img.shape[1] - 1)
            dy = 0.5 * (y[-1] - y[0]) / (img.shape[0] - 1)
            extent = [x[0] - dx, x[-1] + dx,
                      y[-1] + dy, y[0] - dy]

        return ax.imshow(img, extent=extent, **kwargs)


    def _log_base_func(base):
        """
        Convert the base to float and retrieve the simplest function to
        take a logarithm in that base.

        Parameters
        ----------
        base : float
            The base.

        Returns
        -------
        base : float
            The input converted to a :py:class:`float`.
        log : callable
            A function that returns :math:`log_{base}(x)` for any
            array-like `x`.
        """
        base = float(base)
        log = {numpy.e: numpy.log,
               10.0: numpy.log10,
               2.0: numpy.log2}.get(base)

        if log is None:
            lb = numpy.log(base)
            log = lambda x: numpy.log(x) / lb

        return base, log


    def semilogx_base(axes, *args, basex=numpy.e, labelx='e', **kwargs):
        """
        Create a semilogx plot with a custom base.

        The default is to use base `e`.

        This is a convenience wrapper for
        :py:meth:`~matplotlib.axes.Axes.semilogx` which sets ``basex``
        and adds a formatter with a custom label.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            The axes to plot on.
        *args :
            Normal positional arguments to
            :py:meth:`~matplotlib.axes.Axes.semilogx`.
        basex : float, optional
            Base of the x-axis log-scale. The default is `numpy.e`.
        labelx : str, optional
            The label to use to show the log base for tick marks in the
            x-axis. The default is ``'e'``.
        **kwargs :
            Normal keyword arguments to
            :py:meth:`~matplotlib.axes.Axes.semilogx`.

        Returns
        -------
        list of :py:class:`~matplotlib.lines.Line2D`
            Objects representing the plotted data.
        """
        basex, fnx = _log_base_func(basex)

        if labelx is None:
            labelx = str(basex)

        def xticks(x, pos):
            return f'${labelx}^{{{fnx(x):.0f}}}$'

        lines = axes.semilogx(*args, basex=basex, **kwargs)
        axes.xaxis.set_major_formatter(mtick.FuncFormatter(xticks))
        return lines


    def semilogy_base(axes, *args, basey=numpy.e, labely='e', **kwargs):
        """
        Create a semilogy plot with a custom base.

        The default is to use base `e`.

        This is a convenience wrapper for
        :py:meth:`~matplotlib.axes.Axes.semilogy` which sets ``basey``
        and adds a formatter with a custom label.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            The axes to plot on.
        *args :
            Normal positional arguments to
            :py:meth:`~matplotlib.axes.Axes.semilogy`.
        basey : float, optional
            Base of the y-axis log-scale. The default is `numpy.e`.
        labely : str, optional
            The label to use to show the log base for tick marks in the
            y-axis. The default is ``'e'``.
        **kwargs :
            Normal keyword arguments to
            :py:meth:`~matplotlib.axes.Axes.semilogy`.

        Returns
        -------
        list of :py:class:`~matplotlib.lines.Line2D`
            Objects representing the plotted data.
        """
        basey, fny = _log_base_func(basey)

        if labely is None:
            labely = str(basey)

        def yticks(y, pos):
            return f'${labely}^{{{fny(y):.0f}}}$'

        lines = axes.semilogy(*args, basey=basey, **kwargs)
        axes.yaxis.set_major_formatter(mtick.FuncFormatter(yticks))
        return lines


    def loglog_base(axes, *args, basex=numpy.e, basey=numpy.e,
                    labelx='e', labely='e', **kwargs):
        """
        Create a log-log plot with a custom base.

        The default is to use base `e`.

        This is a convenience wrapper for
        :py:meth:`~matplotlib.axes.Axes.loglog` which sets ``basex``,
        ``basey``, and adds a formatter with a custom label.

        Parameters
        ----------
        axes : matplotlib.axes.Axes
            The axes to plot on.
        *args :
            Normal positional arguments to
            :py:meth:`~matplotlib.axes.Axes.loglog`.
        basex : float, optional
            Base of the x-axis log-scale. The default is `numpy.e`.
        basey : float, optional
            Base of the y-axis log-scale. The default is `numpy.e`.
        labelx : str, optional
            The label to use to show the log base for tick marks in the
            x-axis. The default is ``'e'``.
        labely : str, optional
            The label to use to show the log base for tick marks in the
            y-axis. The default is ``'e'``.
        **kwargs :
            Normal keyword arguments to
            :py:meth:`~matplotlib.axes.Axes.loglog`.

        Returns
        -------
        list of :py:class:`~matplotlib.lines.Line2D`
            Objects representing the plotted data.
        """
        basex, fnx = _log_base_func(basex)
        basey, fny = _log_base_func(basey)

        if labelx is None:
            labelx = str(basex)
        if labely is None:
            labely = str(basey)

        def xticks(x, pos):
            return f'${labelx}^{{{fnx(x):.0f}}}$'
        def yticks(y, pos):
            return f'${labely}^{{{fny(y):.0f}}}$'

        lines = axes.loglog(*args, basex=basex, basey=basey, **kwargs)
        axes.xaxis.set_major_formatter(mtick.FuncFormatter(xticks))
        axes.yaxis.set_major_formatter(mtick.FuncFormatter(yticks))
        return lines
